#!/usr/bin/env python

'''
base class for modules generally transfering items using the MISSION_ITEM protocol
'''

import copy
import os
import re
import struct
import time

from pymavlink import mavutil
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_util
if mp_util.has_wxpython:
    from MAVProxy.modules.lib.mp_menu import MPMenuCallFileDialog
    from MAVProxy.modules.lib.mp_menu import MPMenuItem
    from MAVProxy.modules.lib.mp_menu import MPMenuSubMenu

try:
    # py2
    from StringIO import StringIO as SIO
except ImportError:
    # py3
    from io import BytesIO as SIO


class MissionItemProtocolModule(mp_module.MPModule):
    def __init__(self, mpstate, name, description, **args):
        super(MissionItemProtocolModule, self).__init__(mpstate, name, description, **args)
        self.add_command(self.command_name(),
                         self.cmd_wp,
                         '%s management' % self.itemtype(),
                         self.completions())
        self.wp_op = None
        self.wp_requested = {}
        self.wp_received = {}
        self.wp_save_filename = None
        self.wploader_by_sysid = {}
        self.loading_waypoints = False
        self.loading_waypoint_lasttime = time.time()
        self.last_waypoint = 0
        self.wp_period = mavutil.periodic_event(0.5)
        self.undo_wp = None
        self.undo_type = None
        self.undo_wp_idx = -1
        self.upload_start = None
        self.last_get_home = time.time()
        self.ftp_count = None

        if self.continue_mode and self.logdir is not None:
            waytxt = os.path.join(mpstate.status.logdir, self.save_filename())
            if os.path.exists(waytxt):
                self.wploader.load(waytxt)
                print("Loaded %s from %s" % (self.itemstype(), waytxt))

        self.init_gui_menus()

    def gui_menu_items(self):
        return [
            MPMenuItem('Clear', 'Clear', '# %s clear' % self.command_name()),
            MPMenuItem('List', 'List', '# %s list' % self.command_name()),
            MPMenuItem(
                'Load', 'Load', '# %s load ' % self.command_name(),
                handler=MPMenuCallFileDialog(
                    flags=('open',),
                    title='%s Load' % self.mission_type_string(),
                    wildcard='MissionFiles(*.txt.*.wp,*.waypoints)|*.txt;*.wp;*.waypoints')),
            MPMenuItem(
                'Save', 'Save', '# %s save ' % self.command_name(),
                handler=MPMenuCallFileDialog(
                    flags=('save', 'overwrite_prompt'),
                    title='%s Save' % self.mission_type_string(),
                    wildcard='MissionFiles(*.txt.*.wp,*.waypoints)|*.txt;*.wp;*.waypoints')),
            MPMenuItem('Undo', 'Undo', '# %s undo' % self.command_name()),
        ]

    def mission_type_string(self):
        loader_class_string = str(self.loader_class())
        m = re.search("'([^']*)'", loader_class_string)
        if m is None:
            raise ValueError("Failed to match %s" % loader_class_string)
        cname = m.group(1)
        items = cname.split("_")
        return items[-1]

    def init_gui_menus(self):
        '''initialise menus for console and map'''
        self.menu_added_console = False
        self.menu_added_map = False
        if not mp_util.has_wxpython:
            return

        self.menu = MPMenuSubMenu(
            self.mission_type_string(),
            items=self.gui_menu_items()
        )

    def completions(self):
        '''form up MAVProxy-style completion strings used for tab completi
on'''
        cs = self.commands()
        no_arguments = []
        command_argument_buckets = {}
        for c in cs:
            value = cs[c]
            if type(value) == tuple:
                (function, arguments) = value
                args_string = " ".join(arguments)
                if args_string not in command_argument_buckets:
                    command_argument_buckets[args_string] = []
                command_argument_buckets[args_string].append(c)
            else:
                no_arguments.append(c)

        ret = []
        if len(no_arguments):
            ret.append("<" + "|".join(sorted(no_arguments)) + ">")
        for k in command_argument_buckets:
            ret.append("<" + "|".join(sorted(command_argument_buckets[k])) + "> " + k)
        return ret

    def unload(self):
        self.remove_command(self.command_name())
        if self.module('console') is not None and self.menu_added_console:
            self.menu_added_console = False
            self.module('console').remove_menu(self.menu)
        if self.module('map') is not None and self.menu_added_map:
            self.menu_added_map = False
            self.module('map').remove_menu(self.menu)
        super(MissionItemProtocolModule, self).unload()

    def create_loader(self):
        c = self.loader_class()
        return c()

    def last_change(self):
        return self.wploader.last_change

    def check_have_list(self):
        if self.last_change() == 0:
            print("Please list %s items first" % self.command_name())
            return False
        return True

    def index_from_0(self):
        '''e.g. rally points etc are indexed from 1 from the user interface
        perspective'''
        return False

    def save_filename_base(self):
        return self.itemstype().replace(" ", "-")

    def save_filename(self):
        return self.save_filename_base() + ".txt"

    @property
    def wploader(self):
        '''per-sysid wploader'''
        if self.target_system not in self.wploader_by_sysid:
            self.wploader_by_sysid[self.target_system] = self.create_loader()
            self.wploader_by_sysid[self.target_system].expected_count = 0
        return self.wploader_by_sysid[self.target_system]

    def good_item_num_to_manipulate(self, idx):
        if idx > self.wploader.count():
            return False
        if idx < 1:
            return False
        return True

    def item_num_to_offset(self, item_num):
        if self.index_from_0():
            return item_num
        return item_num - 1

    def missing_wps_to_request(self):
        ret = []
        tnow = time.time()
        next_seq = self.wploader.count()
        for i in range(5):
            seq = next_seq+i
            if seq+1 > self.wploader.expected_count:
                continue
            if seq in self.wp_requested and tnow - self.wp_requested[seq] < 2:
                continue
            ret.append(seq)
        return ret

    def append(self, item):
        '''append an item to the held item list'''
        if not self.check_have_list():
            return
        if type(item) == list:
            for i in item:
                self.wploader.add(i)
                self.wploader.expected_count += 1
        else:
            self.wploader.add(item)
            self.wploader.expected_count += 1
        self.wploader.last_change = time.time()
        self.wploader.reindex()

    def send_wp_requests(self, wps=None):
        '''send some more WP requests'''
        if wps is None:
            wps = self.missing_wps_to_request()
        tnow = time.time()
        for seq in wps:
            self.wp_requested[seq] = tnow
            if self.settings.wp_use_mission_int:
                method = self.master.mav.mission_request_int_send
            else:
                method = self.master.mav.mission_request_send
            method(self.target_system,
                   self.target_component,
                   seq,
                   mission_type=self.mav_mission_type())

    def cmd_status(self, args):
        '''show status of wp download'''
        if not self.check_have_list():
            return
        try:
            print("Have %u of %u %s" % (
                self.wploader.count()+len(self.wp_received),
                self.wploader.expected_count,
                self.itemstype()))
        except Exception:
            print("Have %u %s" % (self.wploader.count()+len(self.wp_received), self.itemstype()))

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        mtype = m.get_type()
        if mtype in ['WAYPOINT_COUNT', 'MISSION_COUNT']:
            if getattr(m, 'mission_type', 0) != self.mav_mission_type():
                return
            if self.wp_op is None:
                if self.wploader.expected_count != m.count:
                    self.console.writeln("Mission is stale")
            else:
                self.wploader.clear()
                self.console.writeln("Requesting %u %s t=%s now=%s" % (
                    m.count,
                    self.itemstype(),
                    time.asctime(time.localtime(m._timestamp)),
                    time.asctime()))
                self.wploader.expected_count = m.count
                self.send_wp_requests()

        elif mtype in ['WAYPOINT', 'MISSION_ITEM', 'MISSION_ITEM_INT'] and self.wp_op is not None:
            if m.get_type() == 'MISSION_ITEM_INT':
                if getattr(m, 'mission_type', 0) != self.mav_mission_type():
                    # this is not a mission item, likely fence
                    return
                # our internal structure assumes MISSION_ITEM'''
                m = self.wp_from_mission_item_int(m)
            if m.seq < self.wploader.count():
                # print("DUPLICATE %u" % m.seq)
                return
            if m.seq+1 > self.wploader.expected_count:
                self.console.writeln("Unexpected %s number %u - expected %u" % (self.itemtype(), m.seq, self.wploader.count()))
            self.wp_received[m.seq] = m
            next_seq = self.wploader.count()
            while next_seq in self.wp_received:
                m = self.wp_received.pop(next_seq)
                self.wploader.add(m)
                next_seq += 1
            if self.wploader.count() != self.wploader.expected_count:
                # print("m.seq=%u expected_count=%u" % (m.seq, self.wploader.expected_count))
                self.send_wp_requests()
                return
            if self.wp_op == 'list':
                self.show_and_save(m.get_srcSystem())
                self.loading_waypoints = False
            elif self.wp_op == "save":
                self.save_waypoints(self.wp_save_filename)
            self.wp_op = None
            self.wp_requested = {}
            self.wp_received = {}

        elif mtype in ["WAYPOINT_REQUEST", "MISSION_REQUEST"]:
            self.process_waypoint_request(m, self.master)

    def idle_task(self):
        '''handle missing waypoints'''
        if self.wp_period.trigger():
            # cope with packet loss fetching mission
            if (self.master is not None and
                    self.master.time_since('MISSION_ITEM') >= 2 and
                    self.wploader.count() < getattr(self.wploader, 'expected_count', 0)):
                wps = self.missing_wps_to_request()
                print("re-requesting %s %s" % (self.itemstype(), str(wps)))
                self.send_wp_requests(wps)

        if self.module('console') is not None:
            if not self.menu_added_console:
                self.menu_added_console = True
                self.module('console').add_menu(self.menu)
        else:
            self.menu_added_console = False

        if self.module('map') is not None:
            if not self.menu_added_map:
                self.menu_added_map = True
                self.module('map').add_menu(self.menu)
        else:
            self.menu_added_map = False

    def has_location(self, cmd_id):
        '''return True if a WP command has a location'''
        if cmd_id in mavutil.mavlink.enums['MAV_CMD'].keys():
            cmd_enum = mavutil.mavlink.enums['MAV_CMD'][cmd_id]
            # default to having location for older installs of pymavlink
            # which don't have the attribute
            return getattr(cmd_enum, 'has_location', True)
        return False

    def wp_to_mission_item_int(self, wp):
        '''convert a MISSION_ITEM to a MISSION_ITEM_INT. We always send as
           MISSION_ITEM_INT to give cm level accuracy
        '''
        if wp.get_type() == 'MISSION_ITEM_INT':
            return wp
        if self.has_location(wp.command):
            p5 = int(wp.x*1.0e7)
            p6 = int(wp.y*1.0e7)
        else:
            p5 = int(wp.x)
            p6 = int(wp.y)
        wp_int = mavutil.mavlink.MAVLink_mission_item_int_message(
            wp.target_system,
            wp.target_component,
            wp.seq,
            wp.frame,
            wp.command,
            wp.current,
            wp.autocontinue,
            wp.param1,
            wp.param2,
            wp.param3,
            wp.param4,
            p5,
            p6,
            wp.z,
            wp.mission_type
        )
        return wp_int

    def wp_from_mission_item_int(self, wp):
        '''convert a MISSION_ITEM_INT to a MISSION_ITEM'''
        if self.has_location(wp.command):
            p5 = wp.x*1.0e-7
            p6 = wp.y*1.0e-7
        else:
            p5 = wp.x
            p6 = wp.y
        wp2 = mavutil.mavlink.MAVLink_mission_item_message(wp.target_system,
                                                           wp.target_component,
                                                           wp.seq,
                                                           wp.frame,
                                                           wp.command,
                                                           wp.current,
                                                           wp.autocontinue,
                                                           wp.param1,
                                                           wp.param2,
                                                           wp.param3,
                                                           wp.param4,
                                                           p5,
                                                           p6,
                                                           wp.z,
                                                           wp.mission_type)
        # preserve srcSystem as that is used for naming waypoint file
        wp2._header.srcSystem = wp.get_srcSystem()
        wp2._header.srcComponent = wp.get_srcComponent()
        return wp2

    def process_waypoint_request(self, m, master):
        '''process a waypoint request from the master'''
        if m.mission_type != self.mav_mission_type():
            return
        # print("Processing %s request: (%s)" % (self.itemtype(), str(m)))
        if (m.target_system != self.settings.source_system or
                m.target_component != self.settings.source_component):
            # self.console.error("Mission request is not for me")
            return
        if (not self.loading_waypoints or
                time.time() > self.loading_waypoint_lasttime + 10.0):
            self.loading_waypoints = False
            # self.console.error("not loading waypoints")
            return
        if m.seq >= self.wploader.count():
            self.console.error("Request for bad %s %u (max %u)" %
                               (self.itemtype, m.seq, self.wploader.count()))
            return
        wp = self.wploader.wp(m.seq)
        wp.target_system = self.target_system
        wp.target_component = self.target_component
        if self.settings.wp_use_mission_int:
            wp_send = self.wp_to_mission_item_int(wp)
        else:
            wp_send = wp

        if wp.mission_type != self.mav_mission_type():
            print("Wrong mission type in (%s)" % str(wp))

        self.master.mav.send(wp_send)

        self.loading_waypoint_lasttime = time.time()

        # update the user on our progress:
        self.mpstate.console.set_status(self.itemtype(), '%s %u/%u' % (self.itemtype(), m.seq, self.wploader.count()-1))

        # see if the transfer is complete:
        if m.seq == self.wploader.count() - 1:
            self.loading_waypoints = False
            print("Loaded %u %s in %.2fs" % (
                self.wploader.count(),
                self.itemstype(),
                time.time() - self.upload_start))
            self.console.writeln(
                "Sent all %u %s" %
                (self.wploader.count(), self.itemstype()))

    def send_all_waypoints(self):
        return self.send_all_items()

    def send_all_items(self):
        '''send all waypoints to vehicle'''
        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.upload_start = time.time()
        self.master.mav.mission_count_send(
            self.target_system,
            self.target_component,
            self.wploader.count(),
            mission_type=self.mav_mission_type())

    def load_waypoints(self, filename):
        '''load waypoints from a file'''
        self.wploader.target_system = self.target_system
        self.wploader.target_component = self.target_component
        try:
            # need to remove the leading and trailing quotes in filename
            self.wploader.load(filename.strip('"'))
        except Exception as msg:
            print("Unable to load %s - %s" % (filename, msg))
            return
        print("Loaded %u %s from %s" % (self.wploader.count(), self.itemstype(), filename))
        self.wploader.expected_count = self.wploader.count()
        self.send_all_waypoints()

    def update_waypoints(self, filename, wpnum):
        '''update waypoints from a file'''
        self.wploader.target_system = self.target_system
        self.wploader.target_component = self.target_component
        try:
            self.wploader.load(filename)
        except Exception as msg:
            print("Unable to load %s - %s" % (filename, msg))
            return
        if self.wploader.count() == 0:
            print("No %s found in %s" % (self.itemstype(), filename))
            return
        if wpnum == -1:
            print("Loaded %u updated %s from %s" % (self.wploader.count(), self.itemstype(), filename))
        elif wpnum >= self.wploader.count():
            print("Invalid %s number %u" % (self.itemtype(), wpnum))
            return
        else:
            print("Loaded updated %s %u from %s" % (self.itemtype(), wpnum, filename))

        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        if wpnum == -1:
            start = 0
            end = self.wploader.count()-1
        else:
            start = wpnum
            end = wpnum
        self.master.mav.mission_write_partial_list_send(
            self.target_system,
            self.target_component,
            start,
            end,
            self.mav_mission_type())

    def save_waypoints(self, filename):
        '''save waypoints to a file'''
        try:
            # need to remove the leading and trailing quotes in filename
            self.wploader.save(filename.strip('"'))
        except Exception as msg:
            print("Failed to save %s - %s" % (filename, msg))
            return
        print("Saved %u %s to %s" % (self.wploader.count(), self.itemstype(), filename))

    def save_waypoints_csv(self, filename):
        '''save waypoints to a file in a human readable CSV file'''
        try:
            # need to remove the leading and trailing quotes in filename
            self.wploader.savecsv(filename.strip('"'))
        except Exception as msg:
            print("Failed to save %s - %s" % (filename, msg))
            return
        print("Saved %u %s to CSV %s" % (self.wploader.count(), self.itemstype(), filename))

    def cmd_move(self, args):
        '''handle wp move'''
        if len(args) != 1:
            print("usage: wp move WPNUM")
            return
        idx = int(args[0])
        if not self.good_item_num_to_manipulate(idx):
            print("Invalid %s number %u" % (self.itemtype(), idx))
            return
        offset = self.item_num_to_offset(idx)
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No map click position available")
            return
        wp = self.wploader.wp(offset)

        # setup for undo
        self.undo_wp = copy.copy(wp)
        self.undo_wp_idx = idx
        self.undo_type = "move"

        (lat, lon) = latlon
        if (len(self.module_matching('terrain')) > 0 and
                wp.frame == mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT and
                self.settings.wpterrainadjust):
            alt1 = self.module('terrain').ElevationModel.GetElevation(lat, lon)
            alt2 = self.module('terrain').ElevationModel.GetElevation(wp.x, wp.y)
            if alt1 is not None and alt2 is not None:
                wp.z += alt1 - alt2
        wp.x = lat
        wp.y = lon

        wp.target_system    = self.target_system
        wp.target_component = self.target_component
        self.wploader.set(wp, offset)
        self.wploader.last_change = time.time()

        print("Moving %s %u to %f, %f at %.1fm" % (self.itemtype(), idx, lat, lon, wp.z))

        self.send_single_waypoint(offset)

    def send_single_waypoint(self, idx):
        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.upload_start = time.time()
        self.master.mav.mission_write_partial_list_send(
            self.target_system,
            self.target_component,
            idx,
            idx,
            self.mav_mission_type()
        )

    def is_location_command(self, cmd):
        '''see if cmd is a MAV_CMD with a latitude/longitude'''
        mav_cmd = mavutil.mavlink.enums['MAV_CMD']
        if cmd not in mav_cmd:
            return False
        return getattr(mav_cmd[cmd], 'has_location', True)

    def is_location_wp(self, w):
        '''see if w.command is a MAV_CMD with a latitude/longitude'''
        if w.x == 0 and w.y == 0:
            return False
        return self.is_location_command(w.command)

    def cmd_movemulti(self, args, latlon=None):
        '''handle wp move of multiple waypoints'''
        if len(args) < 3:
            print("usage: wp movemulti WPNUM WPSTART WPEND <rotation>")
            return
        idx = int(args[0])
        if not self.good_item_num_to_manipulate(idx):
            print("Invalid move %s number %u" % (self.itemtype(), idx))
            return
        wpstart = int(args[1])
        if not self.good_item_num_to_manipulate(wpstart):
            print("Invalid start %s number %u" % (self.itemtype(), wpstart))
            return
        wpend = int(args[2])
        if not self.good_item_num_to_manipulate(wpend):
            print("Invalid end %s number %u" % (self.itemtype(), wpend))
            return
        if idx < wpstart or idx > wpend:
            print("WPNUM must be between WPSTART and WPEND")
            return

        # optional rotation about center point
        if len(args) > 3:
            rotation = float(args[3])
        else:
            rotation = 0

        if latlon is None:
            latlon = self.mpstate.click_location
        if latlon is None:
            print("No map click position available")
            return
        idx_offset = self.item_num_to_offset(idx)
        wp = self.wploader.wp(idx_offset)
        if not self.is_location_wp(wp):
            print("WP must be a location command")
            return

        (lat, lon) = latlon
        distance = mp_util.gps_distance(wp.x, wp.y, lat, lon)
        bearing  = mp_util.gps_bearing(wp.x, wp.y, lat, lon)

        wpstart_offset = self.item_num_to_offset(wpstart)
        wpend_offset = self.item_num_to_offset(wpend)
        for wpnum in range(wpstart_offset, wpend_offset+1):
            wp = self.wploader.wp(wpnum)
            if wp is None or not self.is_location_wp(wp):
                continue
            (newlat, newlon) = mp_util.gps_newpos(wp.x, wp.y, bearing, distance)
            if wpnum != idx and rotation != 0:
                # add in rotation
                d2 = mp_util.gps_distance(lat, lon, newlat, newlon)
                b2 = mp_util.gps_bearing(lat, lon, newlat, newlon)
                (newlat, newlon) = mp_util.gps_newpos(lat, lon, b2+rotation, d2)

            if (len(self.module_matching('terrain')) > 0 and
                    wp.frame != mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT and
                    self.settings.wpterrainadjust):
                alt1 = self.module('terrain').ElevationModel.GetElevation(newlat, newlon)
                alt2 = self.module('terrain').ElevationModel.GetElevation(wp.x, wp.y)
                if alt1 is not None and alt2 is not None:
                    wp.z += alt1 - alt2
            wp.x = newlat
            wp.y = newlon
            wp.target_system    = self.target_system
            wp.target_component = self.target_component
            self.wploader.set(wp, wpnum)

        self.wploader.last_change = time.time()
        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.mav.mission_write_partial_list_send(
            self.target_system,
            self.target_component,
            wpstart_offset,
            wpend_offset)
        print("Moved %s %u:%u to %f, %f rotation=%.1f" % (self.itemstype(), wpstart, wpend, lat, lon, rotation))

    def cmd_changealt(self, args):
        '''handle wp change target alt of multiple waypoints'''
        if not self.check_have_list():
            return
        if len(args) < 2:
            print("usage: %s changealt WPNUM NEWALT <NUMWP>" % self.command_name())
            return
        idx = int(args[0])
        if not self.good_item_num_to_manipulate(idx):
            print("Invalid %s number %u" % (self.itemtype(), idx))
            return
        newalt = float(args[1])
        if len(args) >= 3:
            count = int(args[2])
        else:
            count = 1
        if not self.good_item_num_to_manipulate(idx+count-1):
            print("Invalid %s number %u" % (self.itemtype(), idx+count-1))
            return

        for wpnum in range(idx, idx+count):
            offset = self.item_num_to_offset(wpnum)
            wp = self.wploader.wp(offset)
            if not self.wploader.is_location_command(wp.command):
                continue
            wp.z = newalt
            wp.target_system = self.target_system
            wp.target_component = self.target_component
            self.wploader.set(wp, offset)

        self.wploader.last_change = time.time()
        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.mav.mission_write_partial_list_send(
            self.target_system,
            self.target_component,
            offset,
            offset,
            mission_type=self.mav_mission_type())
        print("Changed alt for WPs %u:%u to %f" % (idx, idx+(count-1), newalt))

    def fix_jumps(self, idx, delta):
        # nothing by default as only waypoints need worry
        pass

    def cmd_remove(self, args):
        '''handle wp remove'''
        if len(args) != 1:
            print("usage: %s remove WPNUM" % self.command_name())
            return
        idx = int(args[0])
        if not self.good_item_num_to_manipulate(idx):
            print("Invalid %s number %u" % (self.itemtype(), idx))
            return
        offset = self.item_num_to_offset(idx)
        wp = self.wploader.wp(offset)

        # setup for undo
        self.undo_wp = copy.copy(wp)
        self.undo_wp_idx = idx
        self.undo_type = "remove"

        self.wploader.remove(wp)
        self.wploader.expected_count -= 1
        self.wploader.last_change = time.time()
        self.fix_jumps(offset, -1)
        self.send_all_waypoints()
        print("Removed %s %u" % (self.itemtype(), idx))

    def cmd_undo(self, args):
        '''handle wp undo'''
        if self.undo_wp_idx == -1 or self.undo_wp is None:
            print("No undo information")
            return
        wp = self.undo_wp
        if self.undo_type == 'move':
            wp.target_system    = self.target_system
            wp.target_component = self.target_component
            offset = self.item_num_to_offset(self.undo_wp_idx)
            self.wploader.set(wp, offset)
            self.wploader.last_change = time.time()
            self.send_single_waypoint(offset)
            print("Undid %s move" % self.itemtype())
        elif self.undo_type == 'remove':
            offset = self.item_num_to_offset(self.undo_wp_idx)
            self.wploader.insert(offset, wp)
            self.wploader.expected_count += 1
            self.wploader.last_change = time.time()
            self.fix_jumps(self.undo_wp_idx, 1)
            self.send_all_waypoints()
            print("Undid %s remove" % self.itemtype())
        else:
            print("bad undo type")
        self.undo_wp = None
        self.undo_wp_idx = -1

    def cmd_param(self, args):
        '''handle wp parameter change'''
        if len(args) < 2:
            print("usage: wp param WPNUM PNUM <VALUE>")
            return
        idx = int(args[0])
        if not self.good_item_num_to_manipulate(idx):
            print("Invalid %s number %u" % (self.itemtype(), idx))
            return
        offset = self.item_num_to_offset(idx)
        wp = self.wploader.wp(offset)
        param = [wp.param1, wp.param2, wp.param3, wp.param4]
        pnum = int(args[1])
        if pnum < 1 or pnum > 4:
            print("Invalid param number %u" % pnum)
            return

        if len(args) == 2:
            print("Param %u: %f" % (pnum, param[pnum-1]))
            return

        param[pnum-1] = float(args[2])
        wp.param1 = param[0]
        wp.param2 = param[1]
        wp.param3 = param[2]
        wp.param4 = param[3]

        wp.target_system    = self.target_system
        wp.target_component = self.target_component
        self.wploader.set(wp, idx)
        self.wploader.last_change = time.time()
        self.send_single_waypoint(idx)

    def cmd_clear(self, args):
        self.master.mav.mission_clear_all_send(
            self.target_system,
            self.target_component,
            mission_type=self.mav_mission_type())
        self.wploader.clear()
        if getattr(self.wploader, 'expected_count', None) is not None:
            self.wploader.expected_count = 0
        self.loading_waypoint_lasttime = time.time()

    def cmd_list(self, args):
        self.wp_op = "list"
        self.request_list_send()

    def cmd_load(self, args):
        if len(args) != 1:
            print("usage: %s load FILENAME" % self.command_name())
            return
        self.load_waypoints(args[0])

    def cmd_save(self, args):
        if len(args) != 1:
            print("usage: %s save <filename>" % self.command_name())
            return
        self.wp_save_filename = args[0]
        self.wp_op = "save"
        self.request_list_send()

    def cmd_savecsv(self, args):
        if len(args) != 1:
            print("usage: wp savecsv <filename.csv>")
            return
        self.savecsv(args[0])

    def cmd_savelocal(self, args):
        if len(args) != 1:
            print("usage: wp savelocal <filename>")
            return
        self.wploader.save(args[0])

    def cmd_show(self, args):
        if len(args) != 1:
            print("usage: wp show <filename>")
            return
        self.wploader.load(args[0])

    def cmd_update(self, args):
        if not self.check_have_list():
            return
        if len(args) < 1:
            print("usage: %s update <filename> <wpnum>" % self.command_name())
            return
        if len(args) == 2:
            wpnum = int(args[1])
        else:
            wpnum = -1
        self.update_waypoints(args[0], wpnum)

    def commands(self):
        if self.master and not self.master.mavlink20():
            print("%s module not available; use old compat modules" % str(self.itemtype()))
            return
        return {
            "clear": self.cmd_clear,
            "list": self.cmd_list,
            "load": (self.cmd_load, ["(FILENAME)"]),
            "remove": self.cmd_remove,
            "save": (self.cmd_save, ["(FILENAME)"]),
            "savecsv": (self.cmd_savecsv, ["(FILENAME)"]),
            "savelocal": self.cmd_savelocal,
            "show": (self.cmd_show, ["(FILENAME)"]),
            "status": self.cmd_status,
        }

    def usage(self):
        subcommands = "|".join(sorted(self.commands().keys()))
        return "usage: %s <%s>" % (self.command_name(), subcommands)

    def cmd_wp(self, args):
        '''waypoint commands'''
        if len(args) < 1:
            print(self.usage())
            return

        commands = self.commands()
        if args[0] not in commands:
            print(self.usage())
            return

        function = commands[args[0]]
        if type(function) == tuple:
            (function, function_arguments) = function
            # TODO: do some argument validation here, remove same from
            # cmd_*

        function(args[1:])

    def pretty_enum_value(self, enum_name, enum_value):
        if enum_name == "MAV_FRAME":
            if enum_value == 0:
                return "Abs"
            elif enum_value == 1:
                return "Local"
            elif enum_value == 2:
                return "Mission"
            elif enum_value == 3:
                return "Rel"
            elif enum_value == 4:
                return "Local ENU"
            elif enum_value == 5:
                return "Global (INT)"
            elif enum_value == 10:
                return "AGL"
        ret = mavutil.mavlink.enums[enum_name][enum_value].name
        ret = ret[len(enum_name)+1:]
        return ret

    def csv_line(self, line):
        '''turn a list of values into a CSV line'''
        self.csv_sep = ","
        return self.csv_sep.join(['"' + str(x) + '"' for x in line])

    def pretty_parameter_value(self, value):
        '''pretty parameter value'''
        return value

    def savecsv(self, filename):
        '''save waypoints to a file in human-readable CSV file'''
        f = open(filename, mode='w')
        headers = ["Seq", "Frame", "Cmd", "P1", "P2", "P3", "P4", "X", "Y", "Z"]
        for w in self.wploader.wpoints:
            if getattr(w, 'comment', None):
                #                f.write("# %s\n" % w.comment)
                pass
            out_list = [
                w.seq,
                self.pretty_enum_value('MAV_FRAME', w.frame),
                self.pretty_enum_value('MAV_CMD', w.command),
                self.pretty_parameter_value(w.param1),
                self.pretty_parameter_value(w.param2),
                self.pretty_parameter_value(w.param3),
                self.pretty_parameter_value(w.param4),
                self.pretty_parameter_value(w.x),
                self.pretty_parameter_value(w.y),
                self.pretty_parameter_value(w.z),
            ]
            f.write(self.csv_line(out_list) + "\n")
        f.close()

    def fetch(self):
        """Download wpts from vehicle (this operation is public to support other modules)"""
        if self.wp_op is None:  # If we were already doing a list or save, just restart the fetch without changing the operation  # noqa
            self.wp_op = "fetch"
        self.request_list_send()

    def request_list_send(self):
        self.master.mav.mission_request_list_send(
            self.target_system,
            self.target_component,
            mission_type=self.mav_mission_type())

    def wp_ftp_download(self, args):
        '''Download wpts from vehicle with ftp'''
        ftp = self.mpstate.module('ftp')
        if ftp is None:
            print("Need ftp module")
            return
        self.ftp_count = None
        ftp.cmd_get([self.mission_ftp_name()], callback=self.ftp_callback, callback_progress=self.ftp_callback_progress)

    def ftp_callback_progress(self, fh, total_size):
        '''progress callback from ftp fetch of mission'''
        if self.ftp_count is None and total_size >= 10:
            ofs = fh.tell()
            fh.seek(0)
            buf = fh.read(10)
            fh.seek(ofs)
            magic2, dtype, options, start, num_items = struct.unpack("<HHHHH", buf)
            if magic2 == 0x763d:
                self.ftp_count = num_items
        if self.ftp_count is not None:
            mavmsg = mavutil.mavlink.MAVLink_mission_item_int_message
            item_size = mavmsg.unpacker.size
            done = (total_size - 10) // item_size
            self.mpstate.console.set_status('Mission', 'Mission %u/%u' % (done, self.ftp_count))

    def ftp_callback(self, fh):
        '''callback from ftp fetch of mission'''
        if fh is None:
            print("mission: failed ftp download")
            return
        magic = 0x763d
        data = fh.read()
        magic2, dtype, options, start, num_items = struct.unpack("<HHHHH", data[0:10])
        if magic != magic2:
            print("mission: bad magic 0x%x expected 0x%x" % (magic2, magic))
            return
        if dtype != mavutil.mavlink.MAV_MISSION_TYPE_MISSION:
            print("mission: bad data type %u" % dtype)
            return

        self.wploader.clear()

        data = data[10:]
        mavmsg = mavutil.mavlink.MAVLink_mission_item_int_message
        item_size = mavmsg.unpacker.size
        while len(data) >= item_size:
            mdata = data[:item_size]
            data = data[item_size:]
            msg = mavmsg.unpacker.unpack(mdata)
            tlist = list(msg)
            t = tlist[:]
            for i in range(0, len(tlist)):
                tlist[i] = t[mavmsg.orders[i]]
            t = tuple(tlist)
            w = mavmsg(*t)
            w = self.wp_from_mission_item_int(w)
            self.wploader.add(w)
        self.show_and_save(self.target_system)

    def show_and_save(self, source_system):
        '''display waypoints and save'''
        for i in range(self.wploader.count()):
            w = self.wploader.wp(i)
            print("%u %u %.10f %.10f %f p1=%.1f p2=%.1f p3=%.1f p4=%.1f cur=%u auto=%u" % (
                w.command, w.frame, w.x, w.y, w.z,
                w.param1, w.param2, w.param3, w.param4,
                w.current, w.autocontinue))
        if self.logdir is not None:
            fname = self.save_filename()
            if source_system != 1:
                fname = '%s_%u.txt' % (self.save_filename_base(), source_system)
            waytxt = os.path.join(self.logdir, fname)
            self.save_waypoints(waytxt)
            print("Saved %s to %s" % (self.itemstype(), waytxt))

    def wp_ftp_upload(self, args):
        '''upload waypoints to vehicle with ftp'''
        filename = args[0]
        ftp = self.mpstate.module('ftp')
        if ftp is None:
            print("Need ftp module")
            return
        self.wploader.target_system = self.target_system
        self.wploader.target_component = self.target_component
        try:
            # need to remove the leading and trailing quotes in filename
            self.wploader.load(filename.strip('"'))
        except Exception as msg:
            print("Unable to load %s - %s" % (filename, msg))
            return
        print("Loaded %u waypoints from %s" % (self.wploader.count(), filename))
        print("Sending mission with ftp")

        fh = SIO()
        fh.write(struct.pack("<HHHHH", 0x763d, mavutil.mavlink.MAV_MISSION_TYPE_MISSION, 0, 0, self.wploader.count()))
        mavmsg = mavutil.mavlink.MAVLink_mission_item_int_message
        for i in range(self.wploader.count()):
            w = self.wploader.wp(i)
            w = self.wp_to_mission_item_int(w)
            tlist = []
            for field in mavmsg.ordered_fieldnames:
                tlist.append(getattr(w, field))
            tlist = tuple(tlist)
            buf = mavmsg.unpacker.pack(*tlist)
            fh.write(buf)
        fh.seek(0)

        self.upload_start = time.time()

        ftp.cmd_put([self.mission_ftp_name(), self.mission_ftp_name()],
                    fh=fh, callback=self.ftp_upload_callback, progress_callback=self.ftp_upload_progress)

    def ftp_upload_progress(self, proportion):
        '''callback from ftp put of mission'''
        if proportion is None:
            self.mpstate.console.set_status('Mission', 'Mission ERR')
        else:
            count = self.wploader.count()
            self.mpstate.console.set_status('Mission', 'Mission %u/%u' % (int(proportion*count), count))

    def ftp_upload_callback(self, dlen):
        '''callback from ftp put of mission'''
        if dlen is None:
            print("Failed to send waypoints")
        else:
            mavmsg = mavutil.mavlink.MAVLink_mission_item_int_message
            item_size = mavmsg.unpacker.size
            print("Sent mission of length %u in %.2fs" % ((dlen - 10) // item_size, time.time() - self.upload_start))



#!/usr/bin/env python

"""
  MAVProxy checklist, implemented in a child process
  Created by Stephen Dade (stephen_dade@hotmail.com)
"""

import sys
from MAVProxy.modules.lib import mp_util
from MAVProxy.modules.lib import multiproc
from MAVProxy.modules.lib.wx_loader import wx
from collections import OrderedDict

class CheckItem():
    '''Checklist item used for information transfer
    between threads/processes/pipes'''
    def __init__(self, name, state):
        self.name = name
        self.state = state


class CheckUI():
    '''
    a checklist UI for MAVProxy
    '''
    def __init__(self, title='MAVProxy: Checklist', checklist_file=None):
        import threading
        self.title  = title
        self.menu_callback = None
        self.checklist_file = checklist_file
        self.parent_pipe,self.child_pipe = multiproc.Pipe()
        self.close_event = multiproc.Event()
        self.close_event.clear()
        self.child = multiproc.Process(target=self.child_task)
        self.child.start()

    def child_task(self):
        '''child process - this holds all the GUI elements'''
        mp_util.child_close_fds()
        from MAVProxy.modules.lib.wx_loader import wx

        app = wx.App(False)
        app.frame = ChecklistFrame(state=self, title=self.title)
        app.frame.Show()
        app.MainLoop()

    def close(self):
        '''close the UI'''
        self.close_event.set()
        if self.is_alive():
            self.child.join(2)

    def is_alive(self):
        '''check if child is still going'''
        return self.child.is_alive()

    def set_check(self, name, state):
        '''set a status value'''
        if self.child.is_alive():
            self.parent_pipe.send(CheckItem(name, state))


class ChecklistFrame(wx.Frame):
    """ The main frame of the console"""

    def __init__(self, state, title):
        self.state = state
        wx.Frame.__init__(self, None, title=title, size=(600,600), style=wx.DEFAULT_FRAME_STYLE ^ wx.RESIZE_BORDER)

        #use tabs for the individual checklists
        self.createLists()
        self.panel = wx.Panel(self)
        self.nb = wx.Choicebook(self.panel, wx.ID_ANY)

        #create the tabs
        self.createWidgets()

        #add in the pipe from MAVProxy
        self.timer = wx.Timer(self)
        #self.Bind(wx.EVT_TIMER, self.on_timer, self.timer)
        self.Bind(wx.EVT_TIMER, lambda evt, notebook=self.nb: self.on_timer(evt, notebook), self.timer)
        self.timer.Start(100)

        # finally, put the notebook in a sizer for the panel to manage
        # the layout
        sizer = wx.BoxSizer()
        sizer.Add(self.nb, 1, wx.EXPAND|wx.ALL)
        self.panel.SetSizer(sizer)

        self.Show(True)

    def loadChecklist(self, fname):
        '''load checklist from a file'''
        self.lists = OrderedDict()

        lines = open(fname,'r').readlines()
        for line in lines:
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            a = line.split(':')
            if len(a) < 2:
                continue
            key = a[0].strip()
            text = a[1].strip()
            if not key in self.lists:
                self.lists[key] = []
            self.lists[key].append(text)

    #Create the checklist items
    def createLists(self):
        '''Generate the checklists'''

        if self.state.checklist_file is not None:
            self.loadChecklist(self.state.checklist_file)
            return

        self.lists = OrderedDict()
        self.lists['BeforeAssembly'] = [
            'Confirm batteries charged',
            'No physical damage to airframe',
            'All electronics present and connected',
            'Payload loaded',
            'Ground station operational'
            ]

        self.lists["Takeoff"] = [
            'Engine throttle responsive',
            'Runway clear',
            'Radio links OK',
            'Antenna tracker check',
            'GCS stable'
            ]

    # create controls on form - labels, buttons, etc
    def createWidgets(self):
        #create the panels for the tabs

        for name in self.lists.keys():
            panel = wx.Panel(self.nb)
            box = wx.BoxSizer(wx.VERTICAL)
            panel.SetAutoLayout(True)
            panel.SetSizer(box)

            for key in self.lists[name]:
                CheckBox = wx.CheckBox(panel, wx.ID_ANY, key)
                box.Add(CheckBox)

            panel.Layout()
            self.nb.AddPage(panel, name)

    #Receive messages from MAVProxy and process them
    def on_timer(self, event, notebook):
        state = self.state
        win = notebook.GetPage(notebook.GetSelection())
        if state.close_event.wait(0.001):
            self.timer.Stop()
            self.Destroy()
            return
        while state.child_pipe.poll():
            obj = state.child_pipe.recv()
            if isinstance(obj, CheckItem):
                #go through each item in the current tab and (un)check as needed
                #print(obj.name + ", " + str(obj.state))
                for widget in win.GetChildren():
                    if type(widget) is wx.CheckBox and widget.GetLabel() == obj.name:
                        widget.SetValue(obj.state)


if __name__ == "__main__":
    # test the console
    import time

    checklist = CheckUI()

    #example auto-tick in second tab page
    while checklist.is_alive():
        checklist.set_check("Radio links OK", 1)
        time.sleep(0.5)



#!/usr/bin/env python
'''waypoint command handling'''

from MAVProxy.modules.lib import mission_item_protocol
from MAVProxy.modules.lib import mp_util

from pymavlink import mavutil
from pymavlink import mavwp

import time

if mp_util.has_wxpython:
    from MAVProxy.modules.lib.mp_menu import MPMenuCallTextDialog
    from MAVProxy.modules.lib.mp_menu import MPMenuItem


class WPModule(mission_item_protocol.MissionItemProtocolModule):
    def __init__(self, mpstate):
        super(WPModule, self).__init__(mpstate, "wp", "waypoint handling", public=True)
        # support for setting mission waypoint via command
        self.accepts_DO_SET_MISSION_CURRENT = {}  # keyed by (sysid/compid)

    def gui_menu_items(self):
        ret = super(WPModule, self).gui_menu_items()
        ret.extend([
            MPMenuItem('FTP', 'FTP', '# %s ftp' % self.command_name()),
            MPMenuItem('Editor', 'Editor', '# wp editor'),
            MPMenuItem(
                'Draw', 'Draw', '# wp draw ',
                handler=MPMenuCallTextDialog(
                    title='Mission Altitude (m)',
                    default=100)),
            MPMenuItem('Loop', 'Loop', '# wp loop'),
            MPMenuItem(
                'Add Takeoff', 'Add Takeoff', '# wp add_takeoff ',
                handler=MPMenuCallTextDialog(
                    title='Takeoff Altitude (m)',
                    default=20)),
            MPMenuItem('Add Landing', 'Add Landing', '# wp add_landing'),
            MPMenuItem('Add RTL', 'Add RTL', '# wp add_rtl'),
            MPMenuItem('Add DO_LAND_START', 'Add DO_LAND_START', '# wp add_dls'),
            MPMenuItem('Reset', 'Reset', '# wp set 0'),
        ])
        return ret

    def mission_ftp_name(self):
        return "@MISSION/mission.dat"

    def loader_class(self):
        return mavwp.MAVWPLoader

    def mav_mission_type(self):
        return mavutil.mavlink.MAV_MISSION_TYPE_MISSION

    def save_filename_base(self):
        return 'way'

    def itemstype(self):
        '''returns description of items in the plural'''
        return 'waypoints'

    def itemtype(self):
        '''returns description of item'''
        return 'waypoint'

    def index_from_0(self):
        # other similar user-visible interfaces start indexing
        # user-modifiable items from 1.  waypoints make index 0
        # visible to the user.
        return True

    def command_name(self):
        return "wp"

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        mtype = m.get_type()

        if mtype in ["WAYPOINT_CURRENT", "MISSION_CURRENT"]:
            if m.seq != self.last_waypoint:
                self.last_waypoint = m.seq
                if self.settings.wpupdates:
                    self.say("waypoint %u" % m.seq, priority='message')

        elif mtype == "MISSION_ITEM_REACHED":
            wp = self.wploader.wp(m.seq)
            if wp is None:
                # should we spit out a warning?!
                # self.say("No waypoints")
                pass
            else:
                if wp.command == mavutil.mavlink.MAV_CMD_DO_LAND_START:
                    alt_offset = self.get_mav_param('ALT_OFFSET', 0)
                    if alt_offset > 0.005:
                        self.say("ALT OFFSET IS NOT ZERO passing DO_LAND_START")

        elif mtype == "COMMAND_ACK":
            # check to see if the vehicle has bounced our attempts to
            # set the current mission item via mavlink command (as
            # opposed to the old message):
            if m.command == mavutil.mavlink.MAV_CMD_DO_SET_MISSION_CURRENT:
                key = (m.get_srcSystem(), m.get_srcComponent())
                if m.result == mavutil.mavlink.MAV_RESULT_UNSUPPORTED:
                    # stop sending the commands:
                    self.accepts_DO_SET_MISSION_CURRENT[key] = False
                elif m.result in [mavutil.mavlink.MAV_RESULT_ACCEPTED]:
                    self.accepts_DO_SET_MISSION_CURRENT[key] = True

        super(WPModule, self).mavlink_packet(m)

    def idle_task(self):
        if (self.master is not None and
                'HOME_POSITION' not in self.master.messages and
                time.time() - self.last_get_home > 2):
            self.master.mav.command_long_send(
                self.settings.target_system,
                0,
                mavutil.mavlink.MAV_CMD_GET_HOME_POSITION,
                0, 0, 0, 0, 0, 0, 0, 0)
            self.last_get_home = time.time()

        super(WPModule, self).idle_task()

    def cmd_moverelhome(self, args, latlon=None):
        '''handle wp move to a point relative to home by dist/bearing'''
        if len(args) < 3:
            print("usage: wp moverelhome WPNUM dist bearing")
            return
        idx = int(args[0])
        if idx < 1 or idx > self.wploader.count():
            print("Invalid wp number %u" % idx)
            return
        dist = float(args[1])
        bearing = float(args[2])

        home = self.get_home()
        if home is None:
            print("Need home")
            return

        wp = self.wploader.wp(idx)
        if not self.is_location_wp(wp):
            print("Not a nav command")
            return
        (newlat, newlon) = mp_util.gps_newpos(home.x, home.y, bearing, dist)
        wp.x = newlat
        wp.y = newlon
        wp.target_system    = self.target_system
        wp.target_component = self.target_component
        self.wploader.set(wp, idx)

        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.mav.mission_write_partial_list_send(self.target_system,
                                                        self.target_component,
                                                        idx, idx+1)
        print("Moved WP %u %.1fm bearing %.1f from home" % (idx, dist, bearing))

    def commands(self):
        ret = super(WPModule, self).commands()

        ret.update({
            'add': self.cmd_add,
            "changealt": self.cmd_changealt,
            'draw': self.cmd_draw,
            'editor': self.cmd_editor,
            'loop': self.cmd_loop,
            "param": self.cmd_param,
            "movemulti": self.cmd_movemulti,
            "moverelhome": self.cmd_moverelhome,
            'set': self.cmd_set,
            'sethome': self.cmd_sethome,
            'slope': self.cmd_slope,
            'split': self.cmd_split,
            "move": self.cmd_move,  # handled in parent class
            "add_takeoff": self.wp_add_takeoff,
            "add_landing": self.wp_add_landing,
            "add_rtl": self.wp_add_RTL,
            "add_dls": self.wp_add_dls,
            "ftp": self.wp_ftp_download,
            "ftpload": self.wp_ftp_upload,
            "update": (self.cmd_update, ["(FILENAME)"]),
            "undo": self.cmd_undo,
        })

        return ret

    def mission_type_string(self):
        return 'Mission'

    # waypoint-specific methods:
    def cmd_slope(self, args):
        '''show slope of waypoints'''
        if len(args) == 2:
            # specific waypoints
            wp1 = int(args[0])
            wp2 = int(args[1])
            w1 = self.wploader.wp(wp1)
            w2 = self.wploader.wp(wp2)
            delta_alt = w1.z - w2.z
            if delta_alt == 0:
                slope = "Level"
            else:
                delta_xy = mp_util.gps_distance(w1.x, w1.y, w2.x, w2.y)
                slope = "%.1f" % (delta_xy / delta_alt)
            print("wp%u -> wp%u %s" % (wp1, wp2, slope))
            return
        if len(args) != 0:
            print("Usage: wp slope WP1 WP2")
            return
        last_w = None
        for i in range(1, self.wploader.count()):
            w = self.wploader.wp(i)
            if w.command not in [mavutil.mavlink.MAV_CMD_NAV_WAYPOINT, mavutil.mavlink.MAV_CMD_NAV_LAND]:
                continue
            if last_w is not None:
                if last_w.frame != w.frame:
                    print("WARNING: frame change %u -> %u at %u" % (last_w.frame, w.frame, i))
                delta_alt = last_w.z - w.z
                if delta_alt == 0:
                    slope = "Level"
                else:
                    delta_xy = mp_util.gps_distance(w.x, w.y, last_w.x, last_w.y)
                    slope = "%.1f" % (delta_xy / delta_alt)
                print("WP%u: slope %s" % (i, slope))
            last_w = w

    def get_default_frame(self):
        '''default frame for waypoints'''
        if self.settings.terrainalt == 'Auto':
            if self.get_mav_param('TERRAIN_FOLLOW', 0) == 1:
                return mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT
            return mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT
        if self.settings.terrainalt == 'True':
            return mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT
        return mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT

    def get_home(self):
        '''get home location'''
        if 'HOME_POSITION' in self.master.messages:
            h = self.master.messages['HOME_POSITION']
            return mavutil.mavlink.MAVLink_mission_item_message(self.target_system,
                                                                self.target_component,
                                                                0,
                                                                0,
                                                                mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,
                                                                0, 0, 0, 0, 0, 0,
                                                                h.latitude*1.0e-7, h.longitude*1.0e-7, h.altitude*1.0e-3)
        if self.wploader.count() > 0:
            return self.wploader.wp(0)
        return None

    def wp_draw_callback(self, points):
        '''callback from drawing waypoints'''
        if len(points) < 2:
            return
        self.wploader.target_system = self.target_system
        self.wploader.target_component = self.target_component
        if self.wploader.count() < 2:
            home = self.get_home()
            if home is None:
                print("Need home location for draw - please run gethome")
                return
            self.wploader.clear()
            self.wploader.add(home)
        if self.get_default_frame() == mavutil.mavlink.MAV_FRAME_GLOBAL_TERRAIN_ALT:
            use_terrain = True
        else:
            use_terrain = False
        for p in points:
            self.wploader.add_latlonalt(p[0], p[1], self.settings.wpalt, terrain_alt=use_terrain)
        self.send_all_waypoints()

    def cmd_draw(self, args):
        if 'draw_lines' not in self.mpstate.map_functions:
            print("No map drawing available")
            return
        if self.get_home() is None:
            print("Need home location - please run gethome")
            return
        if len(args) > 1:
            self.settings.wpalt = int(args[1])
        self.mpstate.map_functions['draw_lines'](self.wp_draw_callback)
        print("Drawing %s on map at altitude %d" %
              (self.itemstype(), self.settings.wpalt))

    def cmd_editor(self, args):
        if self.module('misseditor'):
            self.mpstate.functions.process_stdin("module reload misseditor", immediate=True)
        else:
            self.mpstate.functions.process_stdin("module load misseditor", immediate=True)

    def cmd_set(self, args):
        if len(args) != 1:
            print("usage: wp set <wpindex>")
            return

        wp_num = int(args[0])

        # At time of writing MAVProxy sends to (1, 0) by default,
        # but ArduPilot will respond from (1,1) by default -and
        # that means COMMAND_ACK handling will fill
        # self.accepts_DO_SET_MISSION_CURRENT for (1, 1) and we
        # will not get that value here:
        key = (self.target_system, self.target_component)
        supports = self.accepts_DO_SET_MISSION_CURRENT.get(key, None)
        # if we don't know, send both.  If we do know, send only one.
        # we "know" because we hook receipt of COMMAND_ACK.

        if self.settings.wp_use_waypoint_set_current or supports is False:
            self.master.waypoint_set_current_send(wp_num)
        else:
            self.master.mav.command_long_send(
                self.target_system,
                self.target_component,
                mavutil.mavlink.MAV_CMD_DO_SET_MISSION_CURRENT,
                0,
                wp_num, 0, 0, 0, 0, 0, 0
            )

    def cmd_add(self, args):
        '''add a NAV waypoint at the last map click position'''
        if not self.check_have_list():
            return
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No click position available")
            return

        if len(args) < 1:
            alt = self.settings.wpalt
        else:
            alt = float(args[0])

        m = mavutil.mavlink.MAVLink_mission_item_int_message(
            self.target_system,
            self.target_component,
            0,    # seq
            mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,    # frame
            mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,    # command
            0,    # current
            0,    # autocontinue
            0.0,  # param1,
            0.0,  # param2,
            0.0,  # param3
            0.0,  # param4
            int(latlon[0] * 1e7),  # x (latitude)
            int(latlon[1] * 1e7),  # y (longitude)
            alt,                   # z (altitude)
            self.mav_mission_type(),
        )
        self.append(m)
        self.send_all_items()

    def cmd_loop(self, args):
        '''close the loop on a mission'''
        loader = self.wploader
        if loader.count() < 2:
            print("Not enough waypoints (%u)" % loader.count())
            return
        wp = loader.wp(loader.count()-2)
        if wp.command == mavutil.mavlink.MAV_CMD_DO_JUMP:
            print("Mission is already looped")
            return
        if (loader.count() > 1 and
                loader.wp(1).command in [mavutil.mavlink.MAV_CMD_NAV_TAKEOFF, mavutil.mavlink.MAV_CMD_NAV_VTOL_TAKEOFF]):
            target = 2
        else:
            target = 1
        wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0, 0, mavutil.mavlink.MAV_CMD_DO_JUMP,
                                                          0, 1, target, -1, 0, 0, 0, 0, 0)
        loader.add(wp)
        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.waypoint_count_send(self.wploader.count())
        print("Closed loop on mission")

    def is_quadplane(self):
        Q_ENABLE = int(self.get_mav_param("Q_ENABLE", 0))
        return Q_ENABLE > 0

    def wp_add_takeoff(self, args):
        '''add a takeoff as first mission item'''
        if not self.check_have_list():
            return
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No position chosen")
            return
        takeoff_alt = 20
        if len(args) > 0:
            takeoff_alt = float(args[0])
        if self.is_quadplane():
            wptype = mavutil.mavlink.MAV_CMD_NAV_VTOL_TAKEOFF
        else:
            wptype = mavutil.mavlink.MAV_CMD_NAV_TAKEOFF
        wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0,
                                                          mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,
                                                          wptype,
                                                          0, 1, 0, 0, 0, 0, latlon[0], latlon[1], takeoff_alt)
        if self.wploader.count() < 2:
            home = self.get_home()
            if home is None:
                print("Need home location - please run gethome")
                return
            self.wploader.clear()
            self.wploader.add(home)
        # assume first waypoint
        self.wploader.insert(1, wp)
        self.send_all_waypoints()

    def wp_add_landing(self, args):
        '''add a landing as last mission item'''
        if not self.check_have_list():
            return
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No position chosen")
            return
        if self.is_quadplane():
            wptype = mavutil.mavlink.MAV_CMD_NAV_VTOL_LAND
        else:
            wptype = mavutil.mavlink.MAV_CMD_NAV_LAND
        wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0,
                                                          mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,
                                                          wptype,
                                                          0, 1, 0, 0, 0, 0, latlon[0], latlon[1], 0)
        # assume last waypoint
        self.wploader.add(wp)
        self.send_all_waypoints()

    def wp_add_RTL(self, args):
        '''add a RTL as last mission item'''
        if not self.check_have_list():
            return
        wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0,
                                                          mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,
                                                          mavutil.mavlink.MAV_CMD_NAV_RETURN_TO_LAUNCH,
                                                          0, 1, 0, 0, 0, 0, 0, 0, 0)
        # assume last waypoint
        self.wploader.add(wp)
        self.send_all_waypoints()

    def wp_add_dls(self, args):
        '''add a DO_LAND_START as last mission item'''
        if not self.check_have_list():
            return
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No position chosen")
            return
        wp = mavutil.mavlink.MAVLink_mission_item_message(0, 0, 0,
                                                          mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,
                                                          mavutil.mavlink.MAV_CMD_DO_LAND_START,
                                                          0, 1, 0, 0, 0, 0, latlon[0], latlon[1], 0)
        # assume last waypoint
        self.wploader.add(wp)
        self.send_all_waypoints()

    def cmd_sethome(self, args):
        '''set home location from last map click'''
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No position available")
            return
        lat = float(latlon[0])
        lon = float(latlon[1])
        if self.wploader.count() == 0:
            self.wploader.add_latlonalt(lat, lon, 0)
        w = self.wploader.wp(0)
        w.x = lat
        w.y = lon
        self.wploader.set(w, 0)
        self.loading_waypoints = True
        self.loading_waypoint_lasttime = time.time()
        self.master.mav.mission_write_partial_list_send(
            self.target_system,
            self.target_component,
            0, 0
        )

    def fix_jumps(self, idx, delta):
        '''fix up jumps when we add/remove rows'''
        numrows = self.wploader.count()
        for row in range(numrows):
            wp = self.wploader.wp(row)
            jump_cmds = [mavutil.mavlink.MAV_CMD_DO_JUMP]
            if hasattr(mavutil.mavlink, "MAV_CMD_DO_CONDITION_JUMP"):
                jump_cmds.append(mavutil.mavlink.MAV_CMD_DO_CONDITION_JUMP)
            if wp.command in jump_cmds:
                p1 = int(wp.param1)
                if p1 > idx and p1 + delta > 0:
                    wp.param1 = float(p1+delta)
                    self.wploader.set(wp, row)

    def get_loc(self, m):
        '''return a mavutil.location for item m'''
        t = m.get_type()
        if t == "MISSION_ITEM":
            lat = m.x * 1e7
            lng = m.y * 1e7
            alt = m.z * 1e2
        elif t == "MISSION_ITEM_INT":
            lat = m.x
            lng = m.y
            alt = m.z
        else:
            return None
        return mavutil.location(lat, lng, alt)

    def cmd_split(self, args):
        '''splits the segment ended by the supplied waypoint into two'''
        try:
            num = int(args[0])
        except IOError:
            return "Bad wp num (%s)" % args[0]

        if num < 1 or num > self.wploader.count():
            print("Bad item %s" % str(num))
            return
        wp = self.wploader.wp(num)
        if wp is None:
            print("Could not get wp %u" % num)
            return
        loc = self.get_loc(wp)
        if loc is None:
            print("wp is not a location command")
            return

        prev = num - 1
        if prev < 1 or prev > self.wploader.count():
            print("Bad item %u" % num)
            return
        prev_wp = self.wploader.wp(prev)
        if prev_wp is None:
            print("Could not get previous wp %u" % prev)
            return
        prev_loc = self.get_loc(prev_wp)
        if prev_loc is None:
            print("previous wp is not a location command")
            return

        if wp.frame != prev_wp.frame:
            print("waypoints differ in frame (%u vs %u)" %
                  (wp.frame, prev_wp.frame))
            return

        if wp.frame != prev_wp.frame:
            print("waypoints differ in frame")
            return

        lat_avg = (loc.lat + prev_loc.lat)/2
        lng_avg = (loc.lng + prev_loc.lng)/2
        alt_avg = (loc.alt + prev_loc.alt)/2
        new_wp = mavutil.mavlink.MAVLink_mission_item_message(
            self.target_system,
            self.target_component,
            wp.seq,    # seq
            wp.frame,    # frame
            mavutil.mavlink.MAV_CMD_NAV_WAYPOINT,    # command
            0,    # current
            0,    # autocontinue
            0.0,  # param1,
            0.0,  # param2,
            0.0,  # param3
            0.0,  # param4
            lat_avg * 1e-7,  # x (latitude)
            lng_avg * 1e-7,  # y (longitude)
            alt_avg * 1e-2,  # z (altitude)
        )
        self.wploader.insert(wp.seq, new_wp)
        self.wploader.expected_count += 1
        self.fix_jumps(wp.seq, 1)
        self.send_all_waypoints()


def init(mpstate):
    '''initialise module'''
    return WPModule(mpstate)



#!/usr/bin/env python

import lxml.etree as etree
from io import BytesIO as SIO
from zipfile import ZipFile

namespaces = {'kml': 'http://www.opengis.net/kml/2.2'}

def readkmz(filename):
    '''reads in a kmz file and returns xml nodes'''
    #Strip quotation marks if neccessary
    filename.strip('"')
    #Open the zip file (as applicable)    
    if filename[-4:] == '.kml':
        fo = open(filename, "rb")
        fstring = fo.read()
        fo.close()
    elif filename[-4:] == '.kmz':
        zip=ZipFile(filename)
        for z in zip.filelist:
            if z.filename[-4:] == '.kml':
                fstring=zip.read(z)
                break
            else:
                raise Exception("Could not find kml file in %s" % filename)
        else:
            raise Exception("Is not a valid kml or kmz file in %s" % filename)

    parser = etree.XMLParser(encoding='utf-8', recover=True)
    tree = etree.parse(SIO(fstring), parser)
    xpath = xpath = ".//kml:Placemark"

    return tree.findall(xpath, namespaces)

def find_tag(node, tagname):
    for c in node.getchildren():
        if c.tag == "{" + namespaces['kml'] + "}" + tagname:
            return c
    return None

def find_tag_recursive(node, tagname):
    for c in node.getchildren():
        if c.tag == "{" + namespaces['kml'] + "}" + tagname:
            return c
        if hasattr(c, 'getchildren'):
            ret = find_tag_recursive(c, tagname)
            if ret is not None:
                return ret
    return None


def readObject(innode):
    '''reads in a node and returns as a tuple: (type, name, points[])'''
    #get name
    name = find_tag(innode, 'name')
    if name is None:
        return None
    point = find_tag(innode, 'Point')
    if point is not None:
        coordinates = find_tag(point, 'coordinates')
        if coordinates is None:
            return None
        s = coordinates.text.split(',')
        return ("Point", name.text, [(float(s[1]), float(s[0]))])

    coordinates = find_tag_recursive(innode, 'coordinates')
    if coordinates is not None:
        latlon = []
        for c in coordinates.text.split():
            s = c.split(',')
            latlon.append((float(s[1]), float(s[0])))
        return ("Polygon", name.text, latlon)

    return ('Unknown', None, None)

if __name__ == '__main__':
    import sys
    nodes = readkmz(sys.argv[1])
    for n in nodes:
        print(readObject(n))



'''rally point handling via MissionItemProtocol'''

from MAVProxy.modules.lib import mission_item_protocol

from pymavlink import mavutil
from pymavlink import mavwp

from MAVProxy.modules.lib import mp_util

if mp_util.has_wxpython:
    from MAVProxy.modules.lib.mp_menu import MPMenuCallTextDialog
    from MAVProxy.modules.lib.mp_menu import MPMenuItem


class RallyModule(mission_item_protocol.MissionItemProtocolModule):

    def __init__(self, mpstate):
        '''initialise module'''
        super(RallyModule, self).__init__(
            mpstate,
            "rally",
            "rally point management",
            public=True)

    def command_name(self):
        '''command-line command name'''
        return "rally"

    def cmd_rally_add(self, args):
        '''add a rally point at the last map click position'''
        if not self.check_have_list():
            return
        latlon = self.mpstate.click_location
        if latlon is None:
            print("No click position available")
            return

        if len(args) < 1:
            alt = self.settings.rallyalt
        else:
            alt = float(args[0])

        m = mavutil.mavlink.MAVLink_mission_item_int_message(
            self.target_system,
            self.target_component,
            0,    # seq
            mavutil.mavlink.MAV_FRAME_GLOBAL_RELATIVE_ALT,    # frame
            mavutil.mavlink.MAV_CMD_NAV_RALLY_POINT,    # command
            0,    # current
            0,    # autocontinue
            0.0,  # param1,
            0.0,  # param2,
            0.0,  # param3
            0.0,  # param4
            int(latlon[0] * 1e7),  # x (latitude)
            int(latlon[1] * 1e7),  # y (longitude)
            alt,                   # z (altitude)
            self.mav_mission_type(),
        )
        self.append(m)
        self.send_all_items()

    def rally_point(self, i):
        '''return an instance of the old mavlink rally_point message for the
        item at offset i'''
        i = self.wploader.item(i)
        if i is None:
            return None
        lat = i.x
        lng = i.y
        alt = i.z
        if i.get_type() == "MISSION_ITEM":
            lat *= 1e7
            lng *= 1e7
            alt *= 100
        return mavutil.mavlink.MAVLink_rally_point_message(
            i.target_system,
            i.target_component,
            i.seq,
            self.wploader.count(),
            lat,
            lng,
            alt,
            0,
            0,
            0)

    def rally_count(self):
        '''return number of waypoints'''
        return self.wploader.count()

    def commands(self):
        '''returns map from command name to handling function'''
        ret = super(RallyModule, self).commands()
        ret.update({
            'add': self.cmd_rally_add,
            "move": self.cmd_move,  # handled in parent class
            "changealt": self.cmd_changealt,
            "undo": self.cmd_undo,
        })
        return ret

    def mission_ftp_name(self):
        return "@MISSION/fence.dat"

    @staticmethod
    def loader_class():
        return mavwp.MissionItemProtocol_Rally

    def mav_mission_type(self):
        return mavutil.mavlink.MAV_MISSION_TYPE_RALLY

    def itemstype(self):
        '''returns description of items in the plural'''
        return 'rally items'

    def itemtype(self):
        '''returns description of item'''
        return 'rally item'

    def mavlink_packet(self, p):
        super(RallyModule, self).mavlink_packet(p)

    def gui_menu_items(self):
        ret = super(RallyModule, self).gui_menu_items()
        ret.extend([
            MPMenuItem(
                'Add', 'Add', '# rally add ',
                handler=MPMenuCallTextDialog(
                    title='Rally Altitude (m)',
                    default=100
                )
            ),
        ])
        return ret


def init(mpstate):
    '''initialise module'''

    # see if pymavlink is new enough to support new protocols:
    oldmodule = "rallypoint_protocol"
    try:
        mavwp.MissionItemProtocol_Rally
    except AttributeError:
        print("pymavlink too old; using old %s module" % oldmodule)
        mpstate.load_module(oldmodule)
        for (m, pm) in mpstate.modules:
            if m.name == "rally":
                return m
        return None

    return RallyModule(mpstate)



#!/usr/bin/env python
'''
nmea GPS connector
connect to a NMEA GPS on a serial port and provide this as location position
'''

import sys, os, serial, time
from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings
from MAVProxy.modules.lib import mp_util

try:
    import pynmea2
except ImportError as e:
    print('please install pynmea2 package with "sudo apt install python3-nmea2" or "python -m pip install pynmea2"')

class NMEAGPSModule(mp_module.MPModule):
    def __init__(self, mpstate):
        super(NMEAGPSModule, self).__init__(mpstate, "NMEAGPS", "NMEA input")
        self.nmeagps_settings = mp_settings.MPSettings([
            ("port", str, None),
            ("baudrate", int, 9600),
            ])
        self.add_completion_function('(NMEAGPSSETTING)',
                                     self.nmeagps_settings.completion)
        self.add_command('nmeagps', self.cmd_nmeagps, "nmea GPS input control",
                         ["<status|connect|disconnect>", "set (NMEAGPSSETTING)"])
        self.port = None
        self.position = mp_util.mp_position()

    def cmd_nmeagps(self, args):
        '''nmeagps commands'''
        usage = "nmeagps <set|connect|disconnect|status>"
        if len(args) == 0:
            print(usage)
            return
        if args[0] == "set":
            self.nmeagps_settings.command(args[1:])
        elif args[0] == "connect":
            self.cmd_connect()
        elif args[0] == "disconnect":
            self.cmd_disconnect()
        elif args[0] == "status":
            self.cmd_status()
        else:
            print(usage)

    def cmd_connect(self):
        '''connect to GPS'''
        try:
            self.port = serial.Serial(self.nmeagps_settings.port, self.nmeagps_settings.baudrate)
        except Exception as ex:
            print("Failed to open %s : %s" % (self.nmeagps_settings.port, ex))

    def cmd_disconnect(self):
        '''disconnect from GPS'''
        if self.port is not None:
            self.port.close()
            self.port = None
        else:
            print("GPS not connected")

    def cmd_status(self):
        '''status'''
        if self.port is None:
            print("GPS not connected")
            return
        if self.position.timestamp is None:
            print("No position")
            return
        print(self.position)

    def idle_task(self):
        '''check for new data'''
        if self.port is None:
            return
        line = self.port.readline()
        try:
            line = line.decode("ascii")
        except UnicodeDecodeError as e:
            return
        if not line.startswith("$"):
            return
        if line[3:6] == "GGA":
            msg = pynmea2.parse(line)
            self.position.num_sats = int(msg.num_sats)
            self.position.latitude = msg.latitude
            self.position.longitude = msg.longitude
            self.position.altitude = msg.altitude
            self.position.timestamp = time.time()
            self.mpstate.position = self.position
        if line[3:6] == "RMC":
            msg = pynmea2.parse(line)
            if msg.true_course is not None:
                self.position.ground_course = msg.true_course
            else:
                self.position.ground_course = 0.0
            self.position.ground_speed = msg.spd_over_grnd
            self.position.timestamp = time.time()
            self.mpstate.position = self.position

def init(mpstate):
    '''initialise module'''
    return NMEAGPSModule(mpstate)



'''
Support for MAVLink SECURE_COMMAND
'''

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings
from pymavlink import mavutil
import time
import base64
import struct
import random
import glob
try:
    import monocypher
except ImportError:
    print("Please install monocypher with: python3 -m pip install pymonocypher")

class SecureCommandModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(SecureCommandModule, self).__init__(mpstate, "SecureCommand", "SecureCommand Support", public = True)
        self.add_command('securecommand', self.cmd_securecommand, "SecureCommand control",
                         ["<getsessionkey|getpublickeys|setpublickeys|removepublickeys|setconfig>", "set (SECURECOMMANDSETTING)"])

        from MAVProxy.modules.lib.mp_settings import MPSetting
        self.SecureCommand_settings = mp_settings.MPSettings([
            MPSetting("private_keyfile", str, None),
            ])
        self.add_completion_function('(SECURECOMMANDSETTING)',
                                     self.SecureCommand_settings.completion)
        self.session_key = None
        self.public_keys = [None]*10
        self.sequence = random.randint(0, 0xFFFFFFFF)
        self.sent_sequence = None

    def cmd_securecommand(self, args):
        '''securecommand command parser'''
        usage = "usage: securecommand <set|getsessionkey|getpublickeys|setpublickeys|removepublickeys|setconfig>"
        if len(args) == 0:
            print(usage)
            return
        if args[0] == "set":
            self.SecureCommand_settings.command(args[1:])
        elif args[0] == "getsessionkey":
            self.cmd_getsessionkey()
        elif args[0] == "getpublickeys":
            self.cmd_getpublickeys(args[1:])
        elif args[0] == "setpublickeys":
            self.cmd_setpublickeys(args[1:])
        elif args[0] == "removepublickeys":
            self.cmd_removepublickeys(args[1:])
        elif args[0] == "setconfig":
            self.cmd_setconfig(args[1:])
        else:
            print(usage)

    def advance_sequence(self):
        '''add one to sequence'''
        self.sequence = (self.sequence+1) & 0xFFFFFFFF

    def get_private_key(self):
        '''get private key, return 32 byte key or None'''
        if self.SecureCommand_settings.private_keyfile is None:
            return None
        try:
            d = open(self.SecureCommand_settings.private_keyfile,'r').read()
        except Exception as ex:
            return None
        ktype = "PRIVATE_KEYV1:"
        if not d.startswith(ktype):
            return None
        return base64.b64decode(d[len(ktype):])

    def read_public_key(self, keyfile):
        '''read a public key, return 32 byte key or None'''
        try:
            d = open(keyfile,'r').read()
        except Exception as ex:
            return None
        ktype = "PUBLIC_KEYV1:"
        if not d.startswith(ktype):
            return None
        return base64.b64decode(d[len(ktype):])
    
    def have_private_key(self):
        '''return true if we have a valid private key'''
        return self.get_private_key() is not None

    def make_signature(self, seq, command, data):
        '''make a signature'''
        private_key = self.get_private_key()
        d = struct.pack("<II", seq, command)
        d += data
        if command != mavutil.mavlink.SECURE_COMMAND_GET_SESSION_KEY:
            if self.session_key is None:
                print("No session key")
                raise Exception("No session key")
            d += self.session_key
        self.sent_sequence = seq
        return monocypher.signature_sign(private_key, d)

    def pad_data(self, data, dlen=220):
        '''pad data with 0x00 to given length'''
        clen = len(data)
        plen = dlen-clen
        return data + bytearray([0]*plen)

    def cmd_getsessionkey(self):
        '''request session key'''
        if not self.have_private_key():
            print("No private key set")
            return
        sig = self.make_signature(self.sequence, mavutil.mavlink.SECURE_COMMAND_GET_SESSION_KEY, bytes())
        self.master.mav.secure_command_send(self.target_system, self.target_component,
                                            self.sequence, mavutil.mavlink.SECURE_COMMAND_GET_SESSION_KEY,
                                            0, len(sig), self.pad_data(sig))
        self.advance_sequence()

    def cmd_getpublickeys(self, args):
        '''get public keys'''
        if not self.have_private_key():
            print("No private key set")
            return
        if not self.session_key:
            print("No session key")
            return
        idx = 0
        nkeys = 6
        if len(args) > 0:
            idx = int(args[0])
        if len(args) > 1:
            nkeys = int(args[1])
        req = struct.pack("<BB", idx, nkeys)
        sig = self.make_signature(self.sequence, mavutil.mavlink.SECURE_COMMAND_GET_PUBLIC_KEYS, req)
        self.master.mav.secure_command_send(self.target_system, self.target_component,
                                            self.sequence, mavutil.mavlink.SECURE_COMMAND_GET_PUBLIC_KEYS,
                                            len(req), len(sig), self.pad_data(req+sig))
        self.advance_sequence()

    def cmd_removepublickeys(self, args):
        '''remove public keys'''
        if not self.have_private_key():
            print("No private key set")
            return
        if not self.session_key:
            print("No session key")
            return
        if len(args) != 2:
            print("Usage: removepublickeys INDEX COUNT")
            return
        idx = int(args[0])
        nkeys = int(args[1])
        req = struct.pack("<BB", idx, nkeys)
        sig = self.make_signature(self.sequence, mavutil.mavlink.SECURE_COMMAND_REMOVE_PUBLIC_KEYS, req)
        self.master.mav.secure_command_send(self.target_system, self.target_component,
                                            self.sequence, mavutil.mavlink.SECURE_COMMAND_REMOVE_PUBLIC_KEYS,
                                            len(req), len(sig), self.pad_data(req+sig))
        self.advance_sequence()
        
    def cmd_setpublickeys(self, args):
        '''set public keys'''
        if not self.have_private_key():
            print("No private key set")
            return
        if not self.session_key:
            print("No session key")
            return
        if len(args) < 2:
            print("Usage: setpublickeys keyindex KEYFILES...")
            return
        idx = int(args[0])
        keys = []
        for kfile in args[1:]:
            for fname in sorted(glob.glob(kfile)):
                k = self.read_public_key(fname)
                if k is None:
                    print("Unable to load keyfile %s" % fname)
                    return
                print("Loaded key %s" % fname)
                keys.append(k)
        if len(keys) > 6:
            print("Too many keys %u - max is 6" % len(keys))
            return
        if len(keys) == 0:
            print("No keys found")
            return
        req = struct.pack("<B", idx)
        for k in keys:
            req += k
        sig = self.make_signature(self.sequence, mavutil.mavlink.SECURE_COMMAND_SET_PUBLIC_KEYS, req)
        self.master.mav.secure_command_send(self.target_system, self.target_component,
                                            self.sequence, mavutil.mavlink.SECURE_COMMAND_SET_PUBLIC_KEYS,
                                            len(req), len(sig), self.pad_data(req+sig))
        print("Sent %u public keys starting at index %u" % (len(keys), idx))
        self.advance_sequence()

    def cmd_setconfig(self, args):
        '''set configuration parameters'''
        if not self.have_private_key():
            print("No private key set")
            return
        if not self.session_key:
            print("No session key")
            return
        if len(args) < 1:
            print("Usage: setconfig PARAM=VALUE...")
            return
        req = bytearray()
        for i in range(len(args)):
            p = args[i]
            req += p.encode('utf-8')
            if i < len(args)-1:
                req += bytearray([0])
        sig = self.make_signature(self.sequence, mavutil.mavlink.SECURE_COMMAND_SET_REMOTEID_CONFIG, req)
        self.master.mav.secure_command_send(self.target_system, self.target_component,
                                            self.sequence, mavutil.mavlink.SECURE_COMMAND_SET_REMOTEID_CONFIG,
                                            len(req), len(sig), self.pad_data(req+sig))
        print("Sent %u config commands" % len(args))
        self.advance_sequence()
        
    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        if m.get_type() == "SECURE_COMMAND_REPLY":
            if m.sequence != self.sent_sequence:
                print("Invalid reply sequence")
                return
            m.sent_sequence = None

            if m.operation == mavutil.mavlink.SECURE_COMMAND_GET_SESSION_KEY:
                if m.result == mavutil.mavlink.MAV_RESULT_ACCEPTED:
                    self.session_key = bytearray(m.data[:m.data_length])
                    print("Got session key length=%u" % len(self.session_key))
                else:
                    print("Get session key failed: %u" % m.result)

            if m.operation == mavutil.mavlink.SECURE_COMMAND_GET_PUBLIC_KEYS:
                if m.result == mavutil.mavlink.MAV_RESULT_ACCEPTED:
                    idx = m.data[0]
                    if idx >= len(self.public_keys):
                        print("Invalid key index %u" % idx)
                        return
                    keys = bytearray(m.data[1:m.data_length])
                    numkeys = len(keys) // 32
                    if numkeys == 0:
                        print("No public keys returned")
                        return
                    for i in range(idx, numkeys):
                        self.public_keys[i] = keys[32*i:32*(i+1)]
                        keyfile = "public_key%u.dat" % i
                        open(keyfile, "w").write("PUBLIC_KEYV1:" + base64.b64encode(self.public_keys[i]).decode('utf-8') + "\n")
                        print("Wrote %s" % keyfile)
                    print("Got public keys %u to %u" % (idx, numkeys+idx-1))
                else:
                    print("Get public keys failed: %u" % m.result)

            if m.operation == mavutil.mavlink.SECURE_COMMAND_SET_PUBLIC_KEYS:
                if m.result == mavutil.mavlink.MAV_RESULT_ACCEPTED:
                    print("Set public keys OK")
                else:
                    print("Set public keys failed: %u" % m.result)

            if m.operation == mavutil.mavlink.SECURE_COMMAND_REMOVE_PUBLIC_KEYS:
                if m.result == mavutil.mavlink.MAV_RESULT_ACCEPTED:
                    print("Remove public keys OK")
                else:
                    print("Remove public keys failed: %u" % m.result)
                    
def init(mpstate):
    '''initialise module'''
    return SecureCommandModule(mpstate)



'''
Support for OpenDroneID
'''

from MAVProxy.modules.lib import mp_module
from MAVProxy.modules.lib import mp_settings
from pymavlink import mavutil
import time

class OpenDroneIDModule(mp_module.MPModule):

    def __init__(self, mpstate):
        super(OpenDroneIDModule, self).__init__(mpstate, "OpenDroneID", "OpenDroneID Support", public = True)
        self.add_command('opendroneid', self.cmd_opendroneid, "opendroneid control",
                         ["<status>", "set (OPENDRONEIDSETTING)"])

        from MAVProxy.modules.lib.mp_settings import MPSetting
        self.OpenDroneID_settings = mp_settings.MPSettings([
            MPSetting("rate_hz", float, 0.1),
            MPSetting("location_rate_hz", float, 1.0),
            # BASIC_ID
            MPSetting("UAS_ID_type", int, 0, choice=[("None",0), ("SerialNumber",1), ("CAA", 2), ("UTM_ASSIGNED", 3), ("SessionID", 4)]),
            MPSetting("UAS_ID", str, ""),
            MPSetting("UA_type", int, 0, choice=[("None",0), ("Aeroplane",1), ("HeliOrMulti",2),
                                            ("GyroPlane",3), ("HybridLift",4), ("Ornithopter",5),
                                            ("Glider",6), ("Kite",7), ("FreeBalloon",8), ("CaptiveBalloon",9),
                                            ("Airship", 10), ("Parachute",11), ("Rocket",12),
                                            ("TetheredPowered", 13), ("GroundObstacle", 14)]),
            # SELF_ID
            MPSetting("description_type", int, 0, choice=[("Text",0), ("Emergency",1), ("ExtendedStatus", 2)]),
            MPSetting("description", str, ""),
            # SYSTEM
            MPSetting("area_count", int, 1),
            MPSetting("area_radius", int, 0),
            MPSetting("area_ceiling", int, -1000),
            MPSetting("area_floor", int, -1000),
            MPSetting("category_eu", int, 0, choice=[("Undeclared",0), ("Open",1), ("Specific",2), ("Certified",3)]),
            MPSetting("class_eu", int, 0),
            MPSetting("classification_type", int, 0, choice=[("Undeclared",0),("EU",1)]),
            # OPERATOR_ID
            MPSetting("operator_location_type", int, 0, choice=[("Takeoff",0),("LiveGNSS",1),("Fixed",2)]),
            MPSetting("operator_id_type", int, 0),
            MPSetting("operator_id", str, ""),
            ])
        self.add_completion_function('(OPENDRONEIDSETTING)',
                                     self.OpenDroneID_settings.completion)
        self.last_send_s = time.time()
        self.last_loc_send_s = time.time()
        self.next_msg = 0
        self.operator_latitude = 0
        self.operator_longitude = 0
        self.operator_altitude_geo = 0

    def cmd_opendroneid(self, args):
        '''opendroneid command parser'''
        usage = "usage: opendroneid <set>"
        if len(args) == 0:
            print(usage)
            return
        if args[0] == "set":
            self.OpenDroneID_settings.command(args[1:])
        else:
            print(usage)

    def mavlink_packet(self, m):
        '''handle an incoming mavlink packet'''
        pass

    def send_basic_id(self):
        '''send BASIC_ID'''
        self.master.mav.open_drone_id_basic_id_send(
            self.target_system,
            self.target_component,
            self.id_or_mac(),
            self.OpenDroneID_settings.UAS_ID_type,
            self.OpenDroneID_settings.UA_type,
            self.to_bytes(self.OpenDroneID_settings.UAS_ID, 20))

    def send_system(self):
        '''send SYSTEM'''
        if self.mpstate.position is not None:
            pos = self.mpstate.position
            if pos.latitude is not None:
                self.operator_latitude = pos.latitude
                self.operator_longitude = pos.longitude
            if pos.altitude is not None:
                self.operator_altitude_geo = pos.altitude

        self.master.mav.open_drone_id_system_send(
            self.target_system,
            self.target_component,
            self.id_or_mac(),
            self.OpenDroneID_settings.operator_location_type,
            self.OpenDroneID_settings.classification_type,
            int(self.operator_latitude*1.0e7),
            int(self.operator_longitude*1.0e7),
            self.OpenDroneID_settings.area_count,
            self.OpenDroneID_settings.area_radius,
            self.OpenDroneID_settings.area_ceiling,
            self.OpenDroneID_settings.area_floor,
            self.OpenDroneID_settings.category_eu,
            self.OpenDroneID_settings.class_eu,
            self.operator_altitude_geo,
            self.timestamp_2019())

    def send_system_update(self):
        '''send SYSTEM_UPDATE'''
        if self.mpstate.position is not None:
            pos = self.mpstate.position
            if pos.latitude is not None:
                self.operator_latitude = pos.latitude
                self.operator_longitude = pos.longitude
            if pos.altitude is not None:
                self.operator_altitude_geo = pos.altitude

        self.master.mav.open_drone_id_system_update_send(
            self.target_system,
            self.target_component,
            int(self.operator_latitude*1.0e7),
            int(self.operator_longitude*1.0e7),
            self.operator_altitude_geo,
            self.timestamp_2019())
        
    def send_self_id(self):
        '''send SELF_ID'''
        self.master.mav.open_drone_id_self_id_send(
            self.target_system,
            self.target_component,
            self.id_or_mac(),
            self.OpenDroneID_settings.description_type,
            self.to_string(self.OpenDroneID_settings.description, 23))

    def send_operator_id(self):
        '''send OPERATOR_ID'''
        self.master.mav.open_drone_id_operator_id_send(
            self.target_system,
            self.target_component,
            self.id_or_mac(),
            self.OpenDroneID_settings.operator_id_type,
            self.to_string(self.OpenDroneID_settings.operator_id, 20))

    def to_string(self, s, maxlen):
        return s.encode("utf-8")

    def to_bytes(self, s, maxlen):
        b = bytearray(s.encode("utf-8"))
        b = b[:maxlen]
        if len(b) < maxlen:
            b.extend(bytearray([0]*(maxlen-len(b))))
        return b

    def id_or_mac(self):
        return self.to_bytes("", 20)

    def timestamp_2019(self):
        jan_1_2019_s = 1546261200
        return int(time.time() - jan_1_2019_s)

    def idle_task(self):
        '''called on idle'''
        now = time.time()
        if now - self.last_loc_send_s > 1.0/self.OpenDroneID_settings.location_rate_hz:
            self.last_loc_send_s = now
            self.send_system_update()
        if now - self.last_send_s > (1.0/self.OpenDroneID_settings.rate_hz)/4:
            self.last_send_s = now
            if self.next_msg == 0:
                self.send_basic_id()
            elif self.next_msg == 1:
                self.send_system()
            elif self.next_msg == 2:
                self.send_self_id()
            elif self.next_msg == 3:
                self.send_operator_id()
            self.next_msg = (self.next_msg + 1) % 4

def init(mpstate):
    '''initialise module'''
    return OpenDroneIDModule(mpstate)



"""
    MAVProxy geofence module
"""
import copy
import time

from pymavlink import mavextra
from pymavlink import mavutil
from pymavlink import mavwp

from MAVProxy.modules.lib import mission_item_protocol
from MAVProxy.modules.lib import mp_util

if mp_util.has_wxpython:
    from MAVProxy.modules.lib.mp_menu import MPMenuCallTextDialog
    from MAVProxy.modules.lib.mp_menu import MPMenuItem


class FenceModule(mission_item_protocol.MissionItemProtocolModule):
    '''uses common MISSION_ITEM protocol base class to provide fence
    upload/download
    '''

    def __init__(self, mpstate):
        super(FenceModule, self).__init__(mpstate, "fence", "fence point management (new)", public=True)
        self.present = False
        self.enabled = False
        self.healthy = True

    def gui_menu_items(self):
        ret = super(FenceModule, self).gui_menu_items()
        ret.extend([
            MPMenuItem(
                'Add Inclusion Circle', 'Add Inclusion Circle', '# fence addcircle inc ',
                handler=MPMenuCallTextDialog(
                    title='Radius (m)',
                    default=500
                )
            ),
            MPMenuItem(
                'Add Exclusion Circle', 'Add Exclusion Circle', '# fence addcircle exc ',
                handler=MPMenuCallTextDialog(
                    title='Radius (m)',
                    default=500
                )
            ),
            MPMenuItem(
                'Draw Inclusion Poly', 'Draw Inclusion Poly', '# fence draw inc ',
            ),
            MPMenuItem(
                'Draw Exclusion Poly', 'Draw Exclusion Poly', '# fence draw exc ',
            ),
            MPMenuItem(
                'Add Return Point', 'Add Return Point', '# fence addreturnpoint',
            ),
        ])
        return ret

    def command_name(self):
        '''command-line command name'''
        return "fence"

    # def fence_point(self, i):
    #     return self.wploader.fence_point(i)

    def count(self):
        '''return number of waypoints'''
        return self.wploader.count()

    def circles_of_type(self, t):
        '''return a list of Circle fences of a specific type - a single
        MISSION_ITEM'''
        ret = []
        loader = self.wploader
        for i in range(0, loader.count()):
            p = loader.item(i)
            if p is None:
                print("Bad loader item (%u)" % i)
                return []
            if p.command != t:
                continue
            ret.append(p)
        return ret

    def inclusion_circles(self):
        '''return a list of Circle inclusion fences - a single MISSION_ITEM each'''
        return self.circles_of_type(mavutil.mavlink.MAV_CMD_NAV_FENCE_CIRCLE_INCLUSION)

    def exclusion_circles(self):
        '''return a list of Circle exclusion fences - a single MISSION_ITEM each'''
        return self.circles_of_type(mavutil.mavlink.MAV_CMD_NAV_FENCE_CIRCLE_EXCLUSION)

    def polygons_of_type(self, t):
        '''return a list of polygon fences of a specific type - each a list of
        items'''
        ret = []
        loader = self.wploader
        state_outside = 99
        state_inside = 98
        current_polygon = []
        current_expected_length = 0
        state = state_outside
        for i in range(0, loader.count()):
            p = loader.item(i)
            if p is None:
                print("Bad loader item (%u)" % i)
                return []
            if p.command == t:
                # sanity checks:
                if p.param1 < 3:
                    print("Bad vertex count (%u) in seq=%u" % (p.param1, p.seq))
                    continue
                if state == state_outside:
                    # starting a new polygon
                    state = state_inside
                    current_expected_length = p.param1
                    current_polygon = []
                    current_polygon.append(p)
                    continue
                if state == state_inside:
                    # if the count is different and we're in state
                    # inside then the current polygon is invalid.
                    # Discard it.
                    if p.param1 != current_expected_length:
                        print("Short polygon found, discarding")
                        current_expected_length = p.param1
                        current_polygon = []
                        current_polygon.append(p)
                        continue
                    current_polygon.append(p)
                    if len(current_polygon) == current_expected_length:
                        ret.append(current_polygon)
                        state = state_outside
                    continue
                print("Unknown state (%s)" % str(state))
            else:
                if state == state_inside:
                    if len(current_polygon) != current_expected_length:
                        print("Short polygon found")
                    else:
                        ret.append(current_polygon)
                    state = state_outside
                    continue
                if state == state_outside:
                    continue
                print("Unknown state (%s)" % str(state))
        return ret

    def inclusion_polygons(self):
        '''return a list of polygon inclusion fences - each a list of items'''
        return self.polygons_of_type(mavutil.mavlink.MAV_CMD_NAV_FENCE_POLYGON_VERTEX_INCLUSION)

    def exclusion_polygons(self):
        '''return a list of polygon exclusion fences - each a list of items'''
        return self.polygons_of_type(mavutil.mavlink.MAV_CMD_NAV_FENCE_POLYGON_VERTEX_EXCLUSION)

    def returnpoint(self):
        '''return a return point if one exists'''
        loader = self.wploader
        for i in range(0, loader.count()):
            p = loader.item(i)
            if p is None:
                print("Bad loader item (%u)" % i)
                return []
            if p.command != mavutil.mavlink.MAV_CMD_NAV_FENCE_RETURN_POINT:
                continue
            return p
        return None

    @staticmethod
    def loader_class():
        return mavwp.MissionItemProtocol_Fence

    def mav_mission_type(self):
        return mavutil.mavlink.MAV_MISSION_TYPE_FENCE

    def itemstype(self):
        '''returns description of items in the plural'''
        return 'fence items'

    def itemtype(self):
        '''returns description of item'''
        return 'fence item'

    def handle_sys_status(self, m):
        '''function to handle SYS_STATUS packets, used by both old and new module'''
        bits = mavutil.mavlink.MAV_SYS_STATUS_GEOFENCE

        present = ((m.onboard_control_sensors_present & bits) == bits)
        if self.present is False and present is True:
            self.say("fence present")
        elif self.present is True and present is False:
            self.say("fence removed")
        self.present = present

        enabled = ((m.onboard_control_sensors_enabled & bits) == bits)
        if self.enabled is False and enabled is True:
            self.say("fence enabled")
        elif self.enabled is True and enabled is False:
            self.say("fence disabled")
        self.enabled = enabled

        healthy = ((m.onboard_control_sensors_health & bits) == bits)
        if self.healthy is False and healthy is True:
            self.say("fence OK")
        elif self.healthy is True and healthy is False:
            self.say("fence breach")
        self.healthy = healthy

        # console output for fence:
        if not self.present:
            self.console.set_status('Fence', 'FEN', row=0, fg='black')
        elif self.enabled is False:
            self.console.set_status('Fence', 'FEN', row=0, fg='grey')
        elif self.enabled is True and self.healthy is True:
            self.console.set_status('Fence', 'FEN', row=0, fg='green')
        elif self.enabled is True and self.healthy is False:
            self.console.set_status('Fence', 'FEN', row=0, fg='red')

    def mavlink_packet(self, m):
        if m.get_type() == 'SYS_STATUS':
            self.handle_sys_status(m)
        super(FenceModule, self).mavlink_packet(m)

    def fence_draw_callback(self, points):
        '''callback from drawing a fence'''
        if len(points) < 3:
            print("Fence draw cancelled")
            return
        items = []
        for p in points:
            m = mavutil.mavlink.MAVLink_mission_item_int_message(
                self.target_system,
                self.target_component,
                0,    # seq
                mavutil.mavlink.MAV_FRAME_GLOBAL,    # frame
                self.drawing_fence_type,    # command
                0,    # current
                0,    # autocontinue
                len(points), # param1,
                0.0,  # param2,
                0.0,  # param3
                0.0,  # param4
                int(p[0]*1e7),  # x (latitude)
                int(p[1]*1e7),  # y (longitude)
                0,                     # z (altitude)
                self.mav_mission_type(),
            )
            items.append(m)

        self.append(items)
        self.send_all_items()
        self.wploader.last_change = time.time()

    def cmd_draw(self, args):
        '''convenience / compatability / slow learner command to work like the
        old module - i.e. a single inclusion polyfence'''
        # TODO: emit this only if there actually are complex fences:
        if not self.check_have_list():
            return

        if len(args) == 0:
            print("WARNING!  You want 'fence draw inc' or 'fence draw exc'")
            return
        if args[0] in ("inc", "inclusion"):
            self.drawing_fence_type = mavutil.mavlink.MAV_CMD_NAV_FENCE_POLYGON_VERTEX_INCLUSION
            draw_colour = (128, 128, 255)
        elif args[0] in ("exc", "exclusion"):
            self.drawing_fence_type = mavutil.mavlink.MAV_CMD_NAV_FENCE_POLYGON_VERTEX_EXCLUSION
            draw_colour = (255, 128, 128)
        else:
            print("fence draw <inc|inclusion|exc|exclusion>")
            return

        if 'draw_lines' not in self.mpstate.map_functions:
            print("No map drawing available")
            return

        self.mpstate.map_functions['draw_lines'](self.fence_draw_callback,
                                                 colour=draw_colour)
        print("Drawing fence on map")

    def cmd_addcircle(self, args):
        '''adds a circle to the map click position of specific type/radius'''
        if not self.check_have_list():
            return
        if len(args) < 2:
            print("Need 2 arguments")
            return
        t = args[0]
        radius = float(args[1])

        latlon = self.mpstate.click_location
        if latlon is None:
            print("No click position available")
            return

        if t in ["inclusion", "inc"]:
            command = mavutil.mavlink.MAV_CMD_NAV_FENCE_CIRCLE_INCLUSION
        elif t in ["exclusion", "exc"]:
            command = mavutil.mavlink.MAV_CMD_NAV_FENCE_CIRCLE_EXCLUSION
        else:
            print("%s is not one of inclusion|exclusion" % t)
            return

        m = mavutil.mavlink.MAVLink_mission_item_int_message(
            self.target_system,
            self.target_component,
            0,    # seq
            mavutil.mavlink.MAV_FRAME_GLOBAL,    # frame
            command,    # command
            0,    # current
            0,    # autocontinue
            radius, # param1,
            0.0,  # param2,
            0.0,  # param3
            0.0,  # param4
            int(latlon[0] * 1e7),  # x (latitude)
            int(latlon[1] * 1e7),  # y (longitude)
            0,                     # z (altitude)
            self.mav_mission_type(),
        )
        self.append(m)
        self.send_all_items()

    def cmd_addreturnpoint(self, args):
        '''adds a returnpoint at the map click location'''
        if not self.check_have_list():
            return
        latlon = self.mpstate.click_location
        m = mavutil.mavlink.MAVLink_mission_item_int_message(
            self.target_system,
            self.target_component,
            0,    # seq
            mavutil.mavlink.MAV_FRAME_GLOBAL,    # frame
            mavutil.mavlink.MAV_CMD_NAV_FENCE_RETURN_POINT,    # command
            0,    # current
            0,    # autocontinue
            0, # param1,
            0.0,  # param2,
            0.0,  # param3
            0.0,  # param4
            int(latlon[0] * 1e7),  # x (latitude)
            int(latlon[1] * 1e7),  # y (longitude)
            0,                     # z (altitude)
            self.mav_mission_type(),
        )
        self.append(m)
        self.send_all_items()

    def cmd_addpoly(self, args):
        '''adds a number of waypoints equally spaced around a circle around
        click point

        '''
        if not self.check_have_list():
            return
        if len(args) < 1:
            print("Need at least 1 argument (<inclusion|inc|exclusion|exc>", "<radius>" "<pointcount>", "<rotation>")
            return
        t = args[0]
        count = 4
        radius = 20
        rotation = 0
        if len(args) > 1:
            radius = float(args[1])
        if len(args) > 2:
            count = int(args[2])
        if len(args) > 3:
            rotation = float(args[3])

        if count < 3:
            print("Invalid count (%s)" % str(count))
            return
        if radius <= 0:
            print("Invalid radius (%s)" % str(radius))
            return

        latlon = self.mpstate.click_location
        if latlon is None:
            print("No map click position available")
            return

        if t in ["inclusion", "inc"]:
            command = mavutil.mavlink.MAV_CMD_NAV_FENCE_POLYGON_VERTEX_INCLUSION
        elif t in ["exclusion", "exc"]:
            command = mavutil.mavlink.MAV_CMD_NAV_FENCE_POLYGON_VERTEX_EXCLUSION
        else:
            print("%s is not one of inclusion|exclusion" % t)
            return

        items = []
        for i in range(0, count):
            (lat, lon) = mavextra.gps_newpos(latlon[0],
                                             latlon[1],
                                             360/float(count)*i + rotation,
                                             radius)

            m = mavutil.mavlink.MAVLink_mission_item_int_message(
                self.target_system,
                self.target_component,
                0,    # seq
                mavutil.mavlink.MAV_FRAME_GLOBAL,    # frame
                command,    # command
                0,    # current
                0,    # autocontinue
                count, # param1,
                0.0,  # param2,
                0.0,  # param3
                0.0,  # param4
                int(lat*1e7),  # x (latitude)
                int(lon*1e7),  # y (longitude)
                0,                     # z (altitude)
                self.mav_mission_type(),
            )
            items.append(m)

        for m in items:
            self.append(m)
        self.send_all_items()

    def cmd_remove(self, args):
        '''deny remove on fence - requires renumbering etc etc'''
        print("remove is not currently supported for fence.  Try removepolygon_point or removecircle")
        if not self.check_have_list():
            return

    def removereturnpoint(self, seq):
        '''remove returnpoint at offset seq'''
        if not self.check_have_list():
            return
        item = self.wploader.item(seq)
        if item is None:
            print("No item %s" % str(seq))
            return

        if item.command != mavutil.mavlink.MAV_CMD_NAV_FENCE_RETURN_POINT:
            print("Item %u is not a return point" % seq)
            return
        self.wploader.remove(item)
        self.wploader.expected_count -= 1
        self.wploader.last_change = time.time()
        self.send_all_items()

    def removecircle(self, seq):
        '''remove circle at offset seq'''
        if not self.check_have_list():
            return
        item = self.wploader.item(seq)
        if item is None:
            print("No item %s" % str(seq))
            return

        if not self.is_circle_item(item):
            print("Item %u is not a circle" % seq)
            return
        self.wploader.remove(item)
        self.wploader.expected_count -= 1
        self.wploader.last_change = time.time()
        self.send_all_items()

    def is_circle_item(self, item):
        return item.command in [
            mavutil.mavlink.MAV_CMD_NAV_FENCE_CIRCLE_EXCLUSION,
            mavutil.mavlink.MAV_CMD_NAV_FENCE_CIRCLE_INCLUSION,
        ]

    def find_polygon_point(self, polygon_start_seq, item_offset):
        '''find polygon point selected on map
        returns first_item, item_offset
        '''
        first_item = self.wploader.item(polygon_start_seq)
        if first_item is None:
            print("No item at %u" % polygon_start_seq)
            return None, None
        if not self.is_polygon_item(first_item):
            print("Item %u is not a polygon vertex" % polygon_start_seq)
            return None, None
        original_count = int(first_item.param1)
        if item_offset == original_count:
            # selecting closing point on polygon selects first point
            item_offset = 0
        if item_offset > original_count:
            print("Out-of-range point")
            return None, None

        return first_item, item_offset
    
    def removepolygon_point(self, polygon_start_seq, item_offset):
        '''removes item at offset item_offset from the polygon starting at
        polygon_start_seq'''
        if not self.check_have_list():
            return

        items_to_set = []

        first_item, item_offset = self.find_polygon_point(polygon_start_seq, item_offset)
        if first_item is None:
            return
        original_count = int(first_item.param1)
        if original_count <= 3:
            print("Too few points to remove one")
            return

        dead_item_walking = self.wploader.item(polygon_start_seq + item_offset)

        # must reduce count in each of the polygons:
        for i in range(int(first_item.param1)):
            item = self.wploader.item(polygon_start_seq+i)
            if int(item.param1) != original_count:
                print("Invalid polygon starting at %u (count=%u), point %u (count=%u)" %
                      (polygon_start_seq, original_count, i, int(item.param1)))
                return
            item.param1 = item.param1 - 1
            items_to_set.append(item)

        for item in items_to_set:
            w = item
            self.wploader.set(w, w.seq)

        self.wploader.remove(dead_item_walking)
        self.wploader.expected_count -= 1
        self.wploader.last_change = time.time()
        self.send_all_items()

    def addpolygon_point(self, polygon_start_seq, item_offset):
        '''adds item at offset item_offset into the polygon starting at
        polygon_start_seq'''

        if not self.check_have_list():
            return

        items_to_set = []

        first_item = self.wploader.item(polygon_start_seq)
        if (first_item.command not in [
                mavutil.mavlink.MAV_CMD_NAV_FENCE_POLYGON_VERTEX_EXCLUSION,
                mavutil.mavlink.MAV_CMD_NAV_FENCE_POLYGON_VERTEX_INCLUSION
        ]):
            print("Item %u is not a polygon vertex" % polygon_start_seq)
            return
        original_count = int(first_item.param1)
        if item_offset >= original_count:
            print("Out-of-range point")
            return

        # increase count in each of the polygon vertexes:
        for i in range(int(first_item.param1)):
            item = self.wploader.item(polygon_start_seq+i)
            if int(item.param1) != original_count:
                print("Invalid polygon starting at %u (count=%u), point %u (count=%u)" %
                      (polygon_start_seq, original_count, i, int(item.param1)))
                return
            item.param1 = item.param1 + 1
            items_to_set.append(item)

        for item in items_to_set:
            w = item
            self.wploader.set(w, w.seq)

        old_item = self.wploader.item(polygon_start_seq + item_offset)
        new_item = copy.copy(old_item)
        # reset latitude and longitude of new item to be half-way
        # between it and the preceeding point
        if item_offset == 0:
            prev_item_offset = original_count-1
        else:
            prev_item_offset = item_offset - 1
        prev_item = self.wploader.item(polygon_start_seq + prev_item_offset)
        new_item.x = (old_item.x + prev_item.x)/2
        new_item.y = (old_item.y + prev_item.y)/2
        self.wploader.insert(polygon_start_seq + item_offset, new_item)
        self.wploader.expected_count += 1
        self.wploader.last_change = time.time()
        self.send_all_items()

    def is_polygon_item(self, item):
        return item.command in [
            mavutil.mavlink.MAV_CMD_NAV_FENCE_POLYGON_VERTEX_EXCLUSION,
            mavutil.mavlink.MAV_CMD_NAV_FENCE_POLYGON_VERTEX_INCLUSION,
        ]

    def cmd_removepolygon(self, args):
        self.removepolygon(int(args[0]))

    def removepolygon(self, seq):
        '''remove polygon at offset seq'''
        if not self.check_have_list():
            return

        first_item = self.wploader.item(seq)
        if first_item is None:
            print("Invalid item sequence number (%s)" % seq)
            return
        if not self.is_polygon_item(first_item):
            print("Item %u is not a polygon vertex" % seq)
            return

        items_to_remove = []
        for i in range(int(first_item.param1)):
            item = self.wploader.item(seq+i)
            if item is None:
                print("No item %s" % str(i))
                return
            if item.param1 != first_item.param1:
                print("Invalid polygon starting at %u (count=%u), point %u (count=%u)" %
                      (seq, int(first_item.param1), i, int(item.param1)))
                return
            if not self.is_polygon_item(item):
                print("Item %u point %u is not a polygon vertex" % (seq, i))
                return
            items_to_remove.append(item)

        self.wploader.remove(items_to_remove)
        self.wploader.expected_count -= len(items_to_remove)
        self.wploader.last_change = time.time()
        self.send_all_items()

    def cmd_movepolypoint(self, args):
        '''moves item at offset item_offset in polygon starting at
        polygon_start_seqence to map click point'''
        if not self.check_have_list():
            return

        if len(args) < 2:
            print("Need first polygon point and vertex offset")
            return
        polygon_start_seq = int(args[0])
        item_offset = int(args[1])
        first_item, item_offset = self.find_polygon_point(polygon_start_seq, item_offset)
        if first_item is None:
            return

        latlon = self.mpstate.click_location
        if latlon is None:
            print("No map click position available")
            return

        moving_item = self.wploader.item(polygon_start_seq + item_offset)
        moving_item.x = latlon[0]
        moving_item.y = latlon[1]
        if moving_item.get_type() == "MISSION_ITEM_INT":
            moving_item.x *= 1e7
            moving_item.y *= 1e7
            moving_item.x = int(moving_item.x)
            moving_item.y = int(moving_item.y)

        self.wploader.set(moving_item, moving_item.seq)
        self.wploader.last_change = time.time()

        self.send_single_waypoint(moving_item.seq)

    def set_fence_enabled(self, do_enable):
        '''Enable or disable fence'''
        self.master.mav.command_long_send(
            self.target_system,
            self.target_component,
            mavutil.mavlink.MAV_CMD_DO_FENCE_ENABLE,
            0,
            do_enable,
            0,
            0,
            0,
            0,
            0,
            0)

    def cmd_enable(self, args):
        '''enable fence'''
        self.set_fence_enabled(1)

    def cmd_disable(self, args):
        '''disable fence'''
        self.set_fence_enabled(0)

    def commands(self):
        '''returns map from command name to handling function'''
        ret = super(FenceModule, self).commands()
        ret.update({
            'addcircle': (self.cmd_addcircle, ["<inclusion|inc|exclusion|exc>", "RADIUS"]),
            'addpoly': (self.cmd_addpoly, ["<inclusion|inc|exclusion|exc>", "<radius>" "<pointcount>", "<rotation>"]),
            'movepolypoint': (self.cmd_movepolypoint, ["POLY_FIRSTPOINT", "POINT_OFFSET"]),
            'addreturnpoint': (self.cmd_addreturnpoint, []),
            'enable': self.cmd_enable,
            'disable': self.cmd_disable,
            'draw': self.cmd_draw,
            'removepolygon': (self.cmd_removepolygon, ["POLY_FIRSTPOINT"]),
        })
        return ret


def init(mpstate):
    '''initialise module'''
    # see if pymavlink is new enough to support new protocols:
    oldmodule = "fenceitem_protocol"
    try:
        mavwp.MissionItemProtocol_Fence
    except AttributeError:
        print("pymavlink too old; using old %s module" % oldmodule)
        mpstate.load_module(oldmodule)
        for (m, pm) in mpstate.modules:
            if m.name == "fence":
                return m
        return None

    return FenceModule(mpstate)
